from __future__ import annotations

from django.db import models
from django.db.models import Count, Q
from django.utils import timezone
from parler.managers import TranslatableManager, TranslatableQuerySet


class EnhancedImageQuerySet(TranslatableQuerySet):
    def main_images(self):
        return self.filter(is_main=True)

    def secondary_images(self):
        return self.filter(is_main=False)

    def for_product(self, product):
        if hasattr(product, "pk"):
            return self.filter(product=product.pk)
        return self.filter(product=product)

    def for_products(self, product_ids):
        return self.filter(product__in=product_ids)

    def with_titles(self):
        return self.translated().exclude(
            Q(translations__title__isnull=True) | Q(translations__title="")
        )

    def without_titles(self):
        return self.translated().filter(
            Q(translations__title__isnull=True) | Q(translations__title="")
        )

    def recent(self, days=7):
        cutoff = timezone.now() - timezone.timedelta(days=days)
        return self.filter(created_at__gte=cutoff)

    def ordered_by_position(self):
        return self.order_by("sort_order", "created_at")

    def optimized_for_list(self):
        return self.select_related("product").prefetch_related("translations")

    def with_product_data(self):
        return self.select_related("product").prefetch_related(
            "product__translations", "translations"
        )

    def get_products_needing_images(self, max_images=5):
        from product.models.product import Product  # noqa: PLC0415

        return Product.objects.annotate(image_count=Count("images")).filter(
            image_count__lt=max_images
        )

    def get_products_without_main_image(self):
        from django.db.models import Count  # noqa: PLC0415
        from product.models.product import Product  # noqa: PLC0415

        products_with_main_ids = set(
            self.main_images().values_list("product", flat=True)
        )

        products_with_any_images_ids = set(
            self.values_list("product", flat=True).distinct()
        )

        products_without_main_ids = (
            products_with_any_images_ids - products_with_main_ids
        )

        products_no_images_qs = Product.objects.annotate(
            image_count=Count("images")
        ).filter(image_count=0)

        combined_ids = list(products_without_main_ids) + list(
            products_no_images_qs.values_list("id", flat=True)
        )

        return Product.objects.filter(id__in=combined_ids)

    def bulk_update_sort_order(self, image_orders):
        cases = []
        ids = []

        for image_id, sort_order in image_orders.items():
            cases.append(models.When(id=image_id, then=sort_order))
            ids.append(image_id)

        if cases:
            return self.filter(id__in=ids).update(
                sort_order=models.Case(*cases, default=models.F("sort_order"))
            )
        return 0


class EnhancedImageManager(TranslatableManager):
    def get_queryset(self) -> EnhancedImageQuerySet:
        return EnhancedImageQuerySet(self.model, using=self._db)

    def main_image(self, product):
        return self.get_queryset().main_images().for_product(product).first()

    def main_images(self):
        return self.get_queryset().main_images()

    def secondary_images(self):
        return self.get_queryset().secondary_images()

    def for_product(self, product):
        return self.get_queryset().for_product(product)

    def for_products(self, product_ids):
        return self.get_queryset().for_products(product_ids)

    def with_titles(self):
        return self.get_queryset().with_titles()

    def without_titles(self):
        return self.get_queryset().without_titles()

    def recent(self, days=7):
        return self.get_queryset().recent(days)

    def optimized_for_list(self):
        return self.get_queryset().optimized_for_list()

    def with_product_data(self):
        return self.get_queryset().with_product_data()

    def get_products_needing_images(self, max_images=5):
        return self.get_queryset().get_products_needing_images(max_images)

    def get_products_without_main_image(self):
        return self.get_queryset().get_products_without_main_image()

    def bulk_update_sort_order(self, image_orders):
        return self.get_queryset().bulk_update_sort_order(image_orders)

    def ordered_by_position(self):
        return self.get_queryset().ordered_by_position()



# Generated by Django 4.2.9 on 2024-02-10 14:39
import django.db.models.deletion
from django.db import migrations
from django.db import models


class Migration(migrations.Migration):
    initial = True

    dependencies = [
        ("cart", "0001_initial"),
        ("product", "0001_initial"),
    ]

    operations = [
        migrations.AddField(
            model_name="cartitem",
            name="product",
            field=models.ForeignKey(
                on_delete=django.db.models.deletion.CASCADE,
                related_name="cart_item_product",
                to="product.product",
            ),
        ),
    ]



import unittest
from unittest.mock import Mock

from wsgi.health_check import health_check


class TestHealthCheck(unittest.TestCase):
    def setUp(self):
        self.mock_app = Mock()
        self.mock_app.return_value = [b"app response"]
        self.health_url = "/health/"
        self.wrapped_app = health_check(self.mock_app, self.health_url)
        self.mock_start_response = Mock()

    def test_health_check_path(self):
        environ = {"PATH_INFO": self.health_url}
        response = self.wrapped_app(environ, self.mock_start_response)

        self.assertEqual(response, [])

        self.mock_start_response.assert_called_once_with(
            "200 OK", [("Content-Type", "text/plain")]
        )

        self.mock_app.assert_not_called()

    def test_non_health_check_path(self):
        environ = {"PATH_INFO": "/some/other/path/"}
        response = self.wrapped_app(environ, self.mock_start_response)

        self.assertEqual(response, [b"app response"])

        self.mock_app.assert_called_once_with(environ, self.mock_start_response)

        self.mock_start_response.assert_not_called()



from unittest.mock import MagicMock, patch
from urllib.parse import quote

from django.test import TestCase
from rest_framework import status
from rest_framework.exceptions import ValidationError
from rest_framework.test import APIRequestFactory

from blog.models.post import BlogPost, BlogPostTranslation
from product.models.product import Product, ProductTranslation
from search.views import blog_post_meili_search, product_meili_search


class TestBlogPostMeiliSearch(TestCase):
    def setUp(self):
        self.factory = APIRequestFactory()

        self.blog_post = BlogPost.objects.create()

        self.blog_translation = BlogPostTranslation.objects.create(
            master=self.blog_post,
            language_code="en",
            title="Test Blog Post",
            body="This is a test blog post content",
        )

    def test_valid_search_query(self):
        mock_results = {
            "estimated_total_hits": 1,
            "results": [
                {
                    "object": self.blog_translation,
                    "_formatted": {"title": "Test <em>Blog</em> Post"},
                    "_matchesPosition": {"title": [{"start": 5, "length": 4}]},
                    "_rankingScore": 0.95,
                }
            ],
        }

        with patch.object(
            BlogPostTranslation.meilisearch, "paginate"
        ) as mock_paginate:
            mock_paginate.return_value.search.return_value = mock_results

            request = self.factory.get("/search/blog/", {"query": "test"})
            response = blog_post_meili_search(request)

            self.assertEqual(response.status_code, status.HTTP_200_OK)

            data = response.data
            self.assertIn("results", data)
            self.assertIn("estimated_total_hits", data)
            self.assertIn("limit", data)
            self.assertIn("offset", data)

            self.assertEqual(data["limit"], 10)
            self.assertEqual(data["offset"], 0)
            self.assertEqual(data["estimated_total_hits"], 1)

            mock_paginate.assert_called_once_with(limit=10, offset=0)
            mock_paginate.return_value.search.assert_called_once_with(q="test")

    def test_missing_query_parameter(self):
        request = self.factory.get("/search/blog/")

        response = blog_post_meili_search(request)
        self.assertIn("error", response.data)
        self.assertIn(
            "A search query is required.", str(response.data["error"])
        )

    def test_empty_query_parameter(self):
        request = self.factory.get("/search/blog/", {"query": ""})

        try:
            response = blog_post_meili_search(request)
            self.assertIsNotNone(response)
        except ValidationError:
            pass

    def test_custom_pagination_parameters(self):
        mock_results = {"estimated_total_hits": 50, "results": []}

        with patch.object(
            BlogPostTranslation.meilisearch, "paginate"
        ) as mock_paginate:
            mock_paginate.return_value.search.return_value = mock_results

            request = self.factory.get(
                "/search/blog/", {"query": "test", "limit": 20, "offset": 10}
            )
            response = blog_post_meili_search(request)

            self.assertEqual(response.status_code, status.HTTP_200_OK)
            data = response.data
            self.assertEqual(data["limit"], 20)
            self.assertEqual(data["offset"], 10)

            mock_paginate.assert_called_once_with(limit=20, offset=10)

    def test_url_encoded_query(self):
        encoded_query = quote("test query with spaces")
        mock_results = {"estimated_total_hits": 1, "results": []}

        with patch.object(
            BlogPostTranslation.meilisearch, "paginate"
        ) as mock_paginate:
            mock_paginate.return_value.search.return_value = mock_results

            request = self.factory.get(
                "/search/blog/", {"query": encoded_query}
            )
            blog_post_meili_search(request)

            mock_paginate.return_value.search.assert_called_once_with(
                q="test query with spaces"
            )

    def test_search_result_serialization(self):
        mock_results = {
            "estimated_total_hits": 2,
            "results": [
                {
                    "object": self.blog_translation,
                    "_formatted": {"title": "Test <em>Blog</em> Post"},
                    "_matchesPosition": {"title": [{"start": 5, "length": 4}]},
                    "_rankingScore": 0.95,
                },
                {
                    "object": self.blog_translation,
                    "_formatted": {
                        "content": "This is a <em>test</em> content"
                    },
                    "_matchesPosition": {
                        "content": [{"start": 10, "length": 4}]
                    },
                    "_rankingScore": 0.85,
                },
            ],
        }

        with (
            patch.object(
                BlogPostTranslation.meilisearch, "paginate"
            ) as mock_paginate,
            patch(
                "search.views.BlogPostTranslationSerializer"
            ) as mock_serializer,
        ):
            mock_paginate.return_value.search.return_value = mock_results
            mock_serializer.return_value.data = {"title": "Test Blog Post"}

            request = self.factory.get("/search/blog/", {"query": "test"})
            blog_post_meili_search(request)

            self.assertEqual(mock_serializer.call_count, 2)

            for call in mock_serializer.call_args_list:
                args, kwargs = call
                self.assertIn("context", kwargs)
                context = kwargs["context"]
                self.assertIn("_formatted", context)
                self.assertIn("_matchesPosition", context)
                self.assertIn("_rankingScore", context)

    def test_no_search_results(self):
        mock_results = {"estimated_total_hits": 0, "results": []}

        with patch.object(
            BlogPostTranslation.meilisearch, "paginate"
        ) as mock_paginate:
            mock_paginate.return_value.search.return_value = mock_results

            request = self.factory.get(
                "/search/blog/", {"query": "nonexistent"}
            )
            response = blog_post_meili_search(request)

            self.assertEqual(response.status_code, status.HTTP_200_OK)
            data = response.data
            self.assertEqual(data["estimated_total_hits"], 0)
            self.assertEqual(len(data["results"]), 0)

    def test_invalid_pagination_parameters(self):
        mock_results = {"estimated_total_hits": 10, "results": []}

        with patch.object(
            BlogPostTranslation.meilisearch, "paginate"
        ) as mock_paginate:
            mock_paginate.return_value.search.return_value = mock_results

            request = self.factory.get(
                "/search/blog/", {"query": "test", "limit": "invalid"}
            )

            with self.assertRaises(ValueError):
                blog_post_meili_search(request)

    def test_meilisearch_error_handling(self):
        with patch.object(
            BlogPostTranslation.meilisearch, "paginate"
        ) as mock_paginate:
            mock_paginate.side_effect = Exception(
                "MeiliSearch connection error"
            )

            request = self.factory.get("/search/blog/", {"query": "test"})

            with self.assertRaises(Exception) as context:
                blog_post_meili_search(request)

            self.assertIn(
                "MeiliSearch connection error", str(context.exception)
            )


class TestProductMeiliSearch(TestCase):
    def setUp(self):
        self.factory = APIRequestFactory()

        self.product = Product.objects.create()

        self.product_translation = ProductTranslation.objects.create(
            master=self.product,
            language_code="en",
            name="Test Product",
            description="This is a test product description",
        )

    def test_valid_search_query(self):
        mock_results = {
            "estimated_total_hits": 1,
            "results": [
                {
                    "object": self.product_translation,
                    "_formatted": {"name": "Test <em>Product</em>"},
                    "_matchesPosition": {"name": [{"start": 5, "length": 7}]},
                    "_rankingScore": 0.92,
                }
            ],
        }

        with patch.object(
            ProductTranslation.meilisearch, "paginate"
        ) as mock_paginate:
            mock_paginate.return_value.search.return_value = mock_results

            request = self.factory.get("/search/products/", {"query": "test"})
            response = product_meili_search(request)

            self.assertEqual(response.status_code, status.HTTP_200_OK)

            data = response.data
            self.assertIn("results", data)
            self.assertIn("estimated_total_hits", data)
            self.assertIn("limit", data)
            self.assertIn("offset", data)

            self.assertEqual(data["limit"], 10)
            self.assertEqual(data["offset"], 0)
            self.assertEqual(data["estimated_total_hits"], 1)

            mock_paginate.assert_called_once_with(limit=10, offset=0)
            mock_paginate.return_value.search.assert_called_once_with(q="test")

    def test_missing_query_parameter(self):
        request = self.factory.get("/search/products/")

        response = product_meili_search(request)
        self.assertIn("error", response.data)
        self.assertIn(
            "A search query is required.", str(response.data["error"])
        )

    def test_empty_query_parameter(self):
        request = self.factory.get("/search/products/", {"query": ""})

        try:
            response = product_meili_search(request)
            self.assertIsNotNone(response)
        except ValidationError:
            pass

    def test_custom_pagination_parameters(self):
        mock_results = {"estimated_total_hits": 100, "results": []}

        with patch.object(
            ProductTranslation.meilisearch, "paginate"
        ) as mock_paginate:
            mock_paginate.return_value.search.return_value = mock_results

            request = self.factory.get(
                "/search/products/",
                {"query": "laptop", "limit": 25, "offset": 50},
            )
            response = product_meili_search(request)

            self.assertEqual(response.status_code, status.HTTP_200_OK)
            data = response.data
            self.assertEqual(data["limit"], 25)
            self.assertEqual(data["offset"], 50)

            mock_paginate.assert_called_once_with(limit=25, offset=50)

    def test_url_encoded_query(self):
        encoded_query = quote("laptop & accessories")
        mock_results = {"estimated_total_hits": 5, "results": []}

        with patch.object(
            ProductTranslation.meilisearch, "paginate"
        ) as mock_paginate:
            mock_paginate.return_value.search.return_value = mock_results

            request = self.factory.get(
                "/search/products/", {"query": encoded_query}
            )
            product_meili_search(request)

            mock_paginate.return_value.search.assert_called_once_with(
                q="laptop & accessories"
            )

    def test_search_result_serialization(self):
        mock_results = {
            "estimated_total_hits": 3,
            "results": [
                {
                    "object": self.product_translation,
                    "_formatted": {"name": "Test <em>Product</em>"},
                    "_matchesPosition": {"name": [{"start": 5, "length": 7}]},
                    "_rankingScore": 0.92,
                },
                {
                    "object": self.product_translation,
                    "_formatted": {
                        "description": "This is a <em>test</em> description"
                    },
                    "_matchesPosition": {
                        "description": [{"start": 10, "length": 4}]
                    },
                    "_rankingScore": 0.88,
                },
                {
                    "object": self.product_translation,
                    "_formatted": None,
                    "_matchesPosition": None,
                    "_rankingScore": 0.75,
                },
            ],
        }

        with (
            patch.object(
                ProductTranslation.meilisearch, "paginate"
            ) as mock_paginate,
            patch(
                "search.views.ProductTranslationSerializer"
            ) as mock_serializer,
        ):
            mock_paginate.return_value.search.return_value = mock_results
            mock_serializer.return_value.data = {"name": "Test Product"}

            request = self.factory.get("/search/products/", {"query": "test"})
            product_meili_search(request)

            self.assertEqual(mock_serializer.call_count, 3)

            for call in mock_serializer.call_args_list:
                args, kwargs = call
                self.assertIn("context", kwargs)
                context = kwargs["context"]
                self.assertIn("_formatted", context)
                self.assertIn("_matchesPosition", context)
                self.assertIn("_rankingScore", context)

    def test_no_search_results(self):
        mock_results = {"estimated_total_hits": 0, "results": []}

        with patch.object(
            ProductTranslation.meilisearch, "paginate"
        ) as mock_paginate:
            mock_paginate.return_value.search.return_value = mock_results

            request = self.factory.get(
                "/search/products/", {"query": "nonexistent"}
            )
            response = product_meili_search(request)

            self.assertEqual(response.status_code, status.HTTP_200_OK)
            data = response.data
            self.assertEqual(data["estimated_total_hits"], 0)
            self.assertEqual(len(data["results"]), 0)

    def test_large_pagination_values(self):
        mock_results = {"estimated_total_hits": 10000, "results": []}

        with patch.object(
            ProductTranslation.meilisearch, "paginate"
        ) as mock_paginate:
            mock_paginate.return_value.search.return_value = mock_results

            request = self.factory.get(
                "/search/products/",
                {"query": "popular", "limit": 100, "offset": 5000},
            )
            response = product_meili_search(request)

            self.assertEqual(response.status_code, status.HTTP_200_OK)
            data = response.data
            self.assertEqual(data["limit"], 100)
            self.assertEqual(data["offset"], 5000)

            mock_paginate.assert_called_once_with(limit=100, offset=5000)

    def test_meilisearch_error_handling(self):
        with patch.object(
            ProductTranslation.meilisearch, "paginate"
        ) as mock_paginate:
            mock_paginate.side_effect = Exception("MeiliSearch timeout")

            request = self.factory.get("/search/products/", {"query": "test"})

            with self.assertRaises(Exception) as context:
                product_meili_search(request)

            self.assertIn("MeiliSearch timeout", str(context.exception))


class TestSearchViewsCommon(TestCase):
    def setUp(self):
        self.factory = APIRequestFactory()

    def test_query_parameter_extraction(self):
        test_cases = [
            "simple",
            "with spaces",
            "with-special-chars!@#",
            "123numbers",
            "ñoñó unicode",
        ]

        for query in test_cases:
            with self.subTest(query=query):
                request = self.factory.get("/search/blog/", {"query": query})

                self.assertEqual(request.GET.get("query"), query)

    def test_pagination_parameter_defaults(self):
        request = self.factory.get("/search/blog/", {"query": "test"})

        limit = int(request.GET.get("limit", 10))
        offset = int(request.GET.get("offset", 0))

        self.assertEqual(limit, 10)
        self.assertEqual(offset, 0)

    def test_pagination_parameter_conversion(self):
        request = self.factory.get(
            "/search/blog/", {"query": "test", "limit": "20", "offset": "5"}
        )

        limit = int(request.GET.get("limit", 10))
        offset = int(request.GET.get("offset", 0))

        self.assertEqual(limit, 20)
        self.assertEqual(offset, 5)
        self.assertIsInstance(limit, int)
        self.assertIsInstance(offset, int)


class TestSearchViewsEdgeCases(TestCase):
    def setUp(self):
        self.factory = APIRequestFactory()

    def test_zero_pagination_values(self):
        mock_results = {"estimated_total_hits": 100, "results": []}

        with patch.object(
            BlogPostTranslation.meilisearch, "paginate"
        ) as mock_paginate:
            mock_paginate.return_value.search.return_value = mock_results

            request = self.factory.get(
                "/search/blog/", {"query": "test", "limit": 0, "offset": 0}
            )
            response = blog_post_meili_search(request)

            self.assertEqual(response.status_code, status.HTTP_200_OK)
            mock_paginate.assert_called_once_with(limit=0, offset=0)

    def test_negative_pagination_values(self):
        mock_results = {"estimated_total_hits": 100, "results": []}

        with patch.object(
            ProductTranslation.meilisearch, "paginate"
        ) as mock_paginate:
            mock_paginate.return_value.search.return_value = mock_results

            request = self.factory.get(
                "/search/products/",
                {"query": "test", "limit": -5, "offset": -10},
            )
            response = product_meili_search(request)

            self.assertEqual(response.status_code, status.HTTP_200_OK)
            mock_paginate.assert_called_once_with(limit=-5, offset=-10)

    def test_very_long_query(self):
        long_query = "a" * 1000
        mock_results = {"estimated_total_hits": 0, "results": []}

        with patch.object(
            BlogPostTranslation.meilisearch, "paginate"
        ) as mock_paginate:
            mock_paginate.return_value.search.return_value = mock_results

            request = self.factory.get("/search/blog/", {"query": long_query})
            response = blog_post_meili_search(request)

            self.assertEqual(response.status_code, status.HTTP_200_OK)
            mock_paginate.return_value.search.assert_called_once_with(
                q=long_query
            )

    def test_unicode_query(self):
        unicode_query = "héllo wörld 你好 мир"
        mock_results = {"estimated_total_hits": 1, "results": []}

        with patch.object(
            ProductTranslation.meilisearch, "paginate"
        ) as mock_paginate:
            mock_paginate.return_value.search.return_value = mock_results

            request = self.factory.get(
                "/search/products/", {"query": unicode_query}
            )
            response = product_meili_search(request)

            self.assertEqual(response.status_code, status.HTTP_200_OK)
            mock_paginate.return_value.search.assert_called_once_with(
                q=unicode_query
            )

    def test_special_characters_query(self):
        special_query = "!@#$%^&*()[]{}|\\:\";'<>?,./"
        mock_results = {"estimated_total_hits": 0, "results": []}

        with patch.object(
            BlogPostTranslation.meilisearch, "paginate"
        ) as mock_paginate:
            mock_paginate.return_value.search.return_value = mock_results

            request = self.factory.get(
                "/search/blog/", {"query": special_query}
            )
            response = blog_post_meili_search(request)

            self.assertEqual(response.status_code, status.HTTP_200_OK)
            mock_paginate.return_value.search.assert_called_once_with(
                q=special_query
            )

    def test_malformed_meilisearch_results(self):
        malformed_results = {
            "estimated_total_hits": 1,
            "results": [
                {
                    "object": None,
                    "_formatted": {},
                    "_matchesPosition": {},
                    "_rankingScore": None,
                }
            ],
        }

        with (
            patch.object(
                BlogPostTranslation.meilisearch, "paginate"
            ) as mock_paginate,
            patch(
                "search.views.BlogPostTranslationSerializer"
            ) as mock_serializer,
        ):
            mock_paginate.return_value.search.return_value = malformed_results
            mock_serializer.return_value.data = {}

            request = self.factory.get("/search/blog/", {"query": "test"})

            try:
                response = blog_post_meili_search(request)
                self.assertEqual(response.status_code, status.HTTP_200_OK)
            except Exception as e:
                self.assertIsInstance(
                    e, (AttributeError, TypeError, ValueError)
                )

    def test_missing_meilisearch_fields(self):
        results_missing_fields = {
            "estimated_total_hits": 1,
            "results": [
                {
                    "object": MagicMock(),
                }
            ],
        }

        with (
            patch.object(
                ProductTranslation.meilisearch, "paginate"
            ) as mock_paginate,
            patch(
                "search.views.ProductTranslationSerializer"
            ) as mock_serializer,
        ):
            mock_paginate.return_value.search.return_value = (
                results_missing_fields
            )
            mock_serializer.return_value.data = {"name": "Test"}

            request = self.factory.get("/search/products/", {"query": "test"})
            response = product_meili_search(request)

            self.assertEqual(response.status_code, status.HTTP_200_OK)

            mock_serializer.assert_called_once()
            call_kwargs = mock_serializer.call_args[1]
            context = call_kwargs["context"]
            self.assertEqual(context["_formatted"], {})
            self.assertEqual(context["_matchesPosition"], {})
            self.assertIsNone(context["_rankingScore"])



from django.conf import settings
from django.conf.urls.i18n import i18n_patterns
from django.conf.urls.static import static
from django.contrib import admin
from django.contrib.staticfiles.urls import staticfiles_urlpatterns
from django.urls import include, path
from django.utils.translation import gettext_lazy as _
from drf_spectacular.views import (
    SpectacularAPIView,
    SpectacularRedocView,
    SpectacularSwaggerView,
)

from core.api.views import health_check, redirect_to_frontend
from core.views import (
    HomeView,
    ManageTOTPSvgView,
    csp_report,
    robots_txt,
    upload_image,
)
import core.filters.camel_case_filters  # noqa
import core.filters.camel_case_ordering  # noqa

app_name = "core"

urlpatterns = [
    path("robots.txt", robots_txt, name="robots-txt"),
    path("csp_report/", csp_report, name="csp-report"),
    path("i18n/", include("django.conf.urls.i18n")),
]

urlpatterns += i18n_patterns(
    path("", HomeView.as_view(), name="home"),
    path(_("admin/"), admin.site.urls),
    path("upload_image", upload_image, name="upload_image"),
    path("accounts/", include("allauth.urls")),
    path(
        "account/provider/callback",
        redirect_to_frontend,
        name="provider-callback",
    ),
    path("_allauth/", include("allauth.headless.urls")),
    path(
        "_allauth/app/v1/account/authenticators/totp/svg",
        ManageTOTPSvgView.as_api_view(client="app"),
        name="manage_totp_svg",
    ),
    path("rosetta/", include("rosetta.urls")),
    path("tinymce/", include("tinymce.urls")),
    path("api/v1/", include("product.urls")),
    path("api/v1/", include("order.urls")),
    path("api/v1/", include("user.urls")),
    path("api/v1/", include("country.urls")),
    path("api/v1/", include("region.urls")),
    path("api/v1/", include("search.urls")),
    path("api/v1/", include("blog.urls")),
    path("api/v1/", include("tag.urls")),
    path("api/v1/", include("pay_way.urls")),
    path("api/v1/", include("cart.urls")),
    path("api/v1/", include("notification.urls")),
    path("api/v1/", include("contact.urls")),
    path("api/v1/health", health_check, name="api-health"),
    path("api/v1/schema", SpectacularAPIView.as_view(), name="schema"),
    path(
        "api/v1/schema/swagger-ui",
        SpectacularSwaggerView.as_view(url_name="schema"),
        name="swagger-ui",
    ),
    path(
        "api/v1/schema/redoc",
        SpectacularRedocView.as_view(url_name="schema"),
        name="redoc",
    ),
    prefix_default_language=False,
)

if bool(settings.ENABLE_DEBUG_TOOLBAR):
    import warnings

    try:
        import debug_toolbar
    except ImportError:
        warnings.warn(
            "The debug toolbar was not installed. Ignore the error. \
            settings.py should already have warned the user about it.",
            stacklevel=2,
        )
    else:
        urlpatterns += [
            path("__debug__/", include(debug_toolbar.urls)),
        ]

if bool(settings.DEBUG) or settings.SYSTEM_ENV in ["dev", "ci"]:
    urlpatterns += static(
        settings.MEDIA_URL,
        document_root=settings.MEDIA_ROOT,
    )

    urlpatterns += staticfiles_urlpatterns()



from django.contrib.postgres.indexes import BTreeIndex
from django.db import models
from django.utils import timezone
from django.utils.translation import gettext_lazy as _
from django_stubs_ext.db.models import TypedModelMeta

from core.models import TimeStampMixinModel, UUIDModel
from notification.managers import NotificationUserManager


class UnseenNotificationUserManager(models.Manager):
    def get_queryset(self):
        return super().get_queryset().filter(seen=False)


class NotificationUser(TimeStampMixinModel, UUIDModel):
    user = models.ForeignKey(
        "user.UserAccount",
        related_name="notification",
        on_delete=models.CASCADE,
    )
    notification = models.ForeignKey(
        "notification.Notification",
        related_name="user",
        on_delete=models.CASCADE,
    )
    seen = models.BooleanField(_("Seen"), default=False)
    seen_at = models.DateTimeField(_("Seen At"), null=True, blank=True)

    objects: NotificationUserManager = NotificationUserManager()
    unseen_objects: UnseenNotificationUserManager = (
        UnseenNotificationUserManager()
    )

    def __str__(self):
        status = "seen" if self.seen else "unseen"
        return f"Notification {self.notification.id} for {self.user.full_name}: {status}"

    class Meta(TypedModelMeta):
        verbose_name = _("Notification User")
        verbose_name_plural = _("Notification Users")
        ordering = ["-notification__created_at"]
        constraints = [
            models.UniqueConstraint(
                fields=["user", "notification"],
                name="unique_notification_user",
            )
        ]
        indexes = [
            *TimeStampMixinModel.Meta.indexes,
            BTreeIndex(fields=["user"], name="not_user_user_ix"),
            BTreeIndex(fields=["notification"], name="not_user_notif_ix"),
            BTreeIndex(fields=["seen"], name="not_user_seen_ix"),
            BTreeIndex(fields=["seen_at"], name="not_user_seen_at_ix"),
        ]

    def save(self, *args, **kwargs):
        if self.seen and self.seen_at is None:
            self.seen_at = timezone.now()
        super().save(*args, **kwargs)



from datetime import timedelta
from unittest.mock import AsyncMock, MagicMock, patch

from channels.db import database_sync_to_async
from django.contrib.auth import get_user_model
from django.contrib.auth.models import AnonymousUser
from django.test import TransactionTestCase
from django.utils import timezone
from knox.models import get_token_model

from core.middleware.channels import (
    TokenAuthMiddleware,
    TokenAuthMiddlewareStack,
    _cleanup_token,
    _renew_token,
    authenticate_token,
    get_user,
)

User = get_user_model()
Token = get_token_model()


class TestChannelsMiddleware(TransactionTestCase):
    @database_sync_to_async
    def create_test_user(self):
        return User.objects.create_user(
            username="testuser", email="test@test.com", password="password123"
        )

    @database_sync_to_async
    def create_test_token(
        self,
        user,
        token_key="test_token_key",
        digest="test_digest",
        expiry=None,
    ):
        if expiry is None:
            expiry = timezone.now() + timedelta(hours=1)
        elif isinstance(expiry, timedelta):
            expiry = timezone.now() + expiry
        token_instance = Token(
            user=user, token_key=token_key, digest=digest, expiry=expiry
        )
        token_instance.save()
        return token_instance

    async def test_get_user_exists(self):
        user = await self.create_test_user()
        result_user = await get_user(user.id)

        self.assertEqual(result_user.id, user.id)
        self.assertEqual(result_user.username, "testuser")

    async def test_get_user_not_exists(self):
        result_user = await get_user(99999)

        self.assertTrue(result_user.is_anonymous)
        self.assertIsInstance(result_user, AnonymousUser)

    @patch("core.middleware.channels.logger")
    async def test_authenticate_token_invalid_input(self, mock_logger):
        result = await authenticate_token(None)
        self.assertIsInstance(result, AnonymousUser)
        mock_logger.debug.assert_called_with(
            "Token has no strip method, returning AnonymousUser"
        )

        mock_logger.reset_mock()
        result = await authenticate_token(123)
        self.assertIsInstance(result, AnonymousUser)
        mock_logger.debug.assert_called_with(
            "Token has no strip method, returning AnonymousUser"
        )

    @patch("core.middleware.channels.logger")
    async def test_authenticate_token_no_matching_tokens(self, mock_logger):
        result = await authenticate_token("non_existent_token")

        self.assertIsInstance(result, AnonymousUser)
        mock_logger.debug.assert_any_call("Authenticating token: non_existe...")
        mock_logger.debug.assert_any_call(
            "No valid token found, returning AnonymousUser"
        )

    @patch("core.middleware.channels.logger")
    @patch("core.middleware.channels.hash_token")
    async def test_authenticate_token_hash_error(
        self, mock_hash_token, mock_logger
    ):
        user = await self.create_test_user()
        token_string = "test_key123456789abcdef"
        token_key = token_string[:15]
        await self.create_test_token(user, token_key=token_key)

        mock_hash_token.side_effect = TypeError("Hash error")

        result = await authenticate_token(token_string)

        self.assertIsInstance(result, AnonymousUser)
        mock_logger.debug.assert_any_call(f"Found token with key: {token_key}")
        mock_logger.debug.assert_any_call("Error hashing token: Hash error")

    @patch("core.middleware.channels.logger")
    @patch("core.middleware.channels.hash_token")
    @patch("core.middleware.channels.compare_digest")
    async def test_authenticate_token_digest_mismatch(
        self, mock_compare_digest, mock_hash_token, mock_logger
    ):
        user = await self.create_test_user()
        token_string = "test_key123456789abcdef"
        token_key = token_string[:15]
        await self.create_test_token(
            user, token_key=token_key, digest="correct_digest"
        )

        mock_hash_token.return_value = "wrong_digest"
        mock_compare_digest.return_value = False

        result = await authenticate_token(token_string)

        self.assertIsInstance(result, AnonymousUser)
        mock_logger.debug.assert_any_call("Token digest mismatch")

    @patch("core.middleware.channels.logger")
    @patch("core.middleware.channels.hash_token")
    @patch("core.middleware.channels.compare_digest")
    async def test_authenticate_token_success(
        self, mock_compare_digest, mock_hash_token, mock_logger
    ):
        user = await self.create_test_user()
        token_string = "test_key123456789abcdef"
        token_key = token_string[:15]
        await self.create_test_token(
            user, token_key=token_key, digest="correct_digest"
        )

        mock_hash_token.return_value = "correct_digest"
        mock_compare_digest.return_value = True

        result = await authenticate_token(token_string)

        self.assertEqual(result.id, user.id)
        mock_logger.debug.assert_any_call("Token digest match")
        mock_logger.debug.assert_any_call(
            f"Authenticated user: {user.username}"
        )

    @patch("core.middleware.channels.logger")
    @patch("core.middleware.channels.hash_token")
    @patch("core.middleware.channels.compare_digest")
    async def test_authenticate_token_inactive_user(
        self, mock_compare_digest, mock_hash_token, mock_logger
    ):
        user = await self.create_test_user()
        user.is_active = False
        await database_sync_to_async(user.save)()

        token_string = "test_key123456789abcdef"
        token_key = token_string[:15]
        await self.create_test_token(
            user, token_key=token_key, digest="correct_digest"
        )

        mock_hash_token.return_value = "correct_digest"
        mock_compare_digest.return_value = True

        result = await authenticate_token(token_string)

        self.assertIsInstance(result, AnonymousUser)
        mock_logger.debug.assert_any_call("Token digest match")
        mock_logger.debug.assert_any_call("Authenticated user: AnonymousUser")

    @patch("core.middleware.channels.logger")
    @patch("core.middleware.channels.knox_settings")
    async def test_authenticate_token_with_auto_refresh(
        self, mock_knox_settings, mock_logger
    ):
        mock_knox_settings.AUTO_REFRESH = True

        user = await self.create_test_user()
        token_string = "test_key123456789abcdef"
        token_key = token_string[:15]
        expiry_time = timezone.now() + timedelta(hours=1)
        token = await self.create_test_token(
            user,
            token_key=token_key,
            digest="correct_digest",
            expiry=expiry_time,
        )

        with (
            patch("core.middleware.channels.hash_token") as mock_hash_token,
            patch(
                "core.middleware.channels.compare_digest"
            ) as mock_compare_digest,
            patch("core.middleware.channels._renew_token") as mock_renew_token,
        ):
            mock_hash_token.return_value = "correct_digest"
            mock_compare_digest.return_value = True

            result = await authenticate_token(token_string)

            self.assertEqual(result.id, user.id)
            mock_logger.debug.assert_any_call("Auto refreshing token")
            mock_renew_token.assert_called_once_with(token)

    async def test_cleanup_token_expired(self):
        user = await self.create_test_user()
        expired_time = timezone.now() - timedelta(hours=1)
        token = await self.create_test_token(user, expiry=expired_time)

        result = await database_sync_to_async(_cleanup_token)(token)

        self.assertTrue(result)
        with self.assertRaises(Token.DoesNotExist):
            await database_sync_to_async(Token.objects.get)(pk=token.pk)

    async def test_cleanup_token_not_expired(self):
        user = await self.create_test_user()
        future_time = timezone.now() + timedelta(hours=1)
        token = await self.create_test_token(user, expiry=future_time)

        result = await database_sync_to_async(_cleanup_token)(token)

        self.assertFalse(result)
        existing_token = await database_sync_to_async(Token.objects.get)(
            pk=token.pk
        )
        self.assertEqual(existing_token.pk, token.pk)

    async def test_cleanup_token_no_expiry(self):
        user = await self.create_test_user()
        token = await self.create_test_token(user, expiry=None)

        result = await database_sync_to_async(_cleanup_token)(token)

        self.assertFalse(result)
        existing_token = await database_sync_to_async(Token.objects.get)(
            pk=token.pk
        )
        self.assertEqual(existing_token.pk, token.pk)

    @patch("core.middleware.channels.logger")
    @patch("core.middleware.channels.knox_settings")
    async def test_renew_token_significant_delta(
        self, mock_knox_settings, mock_logger
    ):
        mock_knox_settings.TOKEN_TTL = timedelta(hours=24)
        mock_knox_settings.MIN_REFRESH_INTERVAL = 300

        user = await self.create_test_user()
        old_expiry = timezone.now() + timedelta(hours=1)
        token = await self.create_test_token(user, expiry=old_expiry)

        await database_sync_to_async(_renew_token)(token)

        await database_sync_to_async(token.refresh_from_db)()

        self.assertGreater(token.expiry, old_expiry)
        mock_logger.debug.assert_called()

    @patch("core.middleware.channels.logger")
    @patch("core.middleware.channels.knox_settings")
    async def test_renew_token_small_delta(
        self, mock_knox_settings, mock_logger
    ):
        mock_knox_settings.TOKEN_TTL = timedelta(seconds=1)
        mock_knox_settings.MIN_REFRESH_INTERVAL = 300

        user = await self.create_test_user()
        old_expiry = timezone.now() + timedelta(hours=1)
        token = await self.create_test_token(user, expiry=old_expiry)

        await database_sync_to_async(_renew_token)(token)

        await database_sync_to_async(token.refresh_from_db)()

        self.assertEqual(token.expiry, old_expiry)

    @patch("core.middleware.channels.logger")
    async def test_token_auth_middleware_with_access_token(self, mock_logger):
        async def mock_inner(scope, receive, send):
            pass

        middleware = TokenAuthMiddleware(mock_inner)

        scope = {
            "query_string": b"access_token=test_access_token&other_param=value",
            "user": None,
        }

        with patch("core.middleware.channels.authenticate_token") as mock_auth:
            mock_user = MagicMock()
            mock_user.is_anonymous = False
            mock_user.username = "testuser"
            mock_auth.return_value = mock_user

            await middleware(scope, AsyncMock(), AsyncMock())

            mock_auth.assert_called_once_with("test_access_token")
            self.assertEqual(scope["user"], mock_user)
            mock_logger.debug.assert_any_call(
                "TokenAuthMiddleware Token found, authenticating"
            )
            mock_logger.debug.assert_any_call("User authenticated: testuser")

    @patch("core.middleware.channels.logger")
    async def test_token_auth_middleware_with_session_token(self, mock_logger):
        async def mock_inner(scope, receive, send):
            pass

        middleware = TokenAuthMiddleware(mock_inner)

        scope = {
            "query_string": b"session_token=test_session_token",
            "user": None,
        }

        with patch("core.middleware.channels.authenticate_token") as mock_auth:
            mock_user = MagicMock()
            mock_user.is_anonymous = False
            mock_user.username = "testuser"
            mock_auth.return_value = mock_user

            await middleware(scope, AsyncMock(), AsyncMock())

            mock_auth.assert_called_once_with("test_session_token")
            self.assertEqual(scope["user"], mock_user)

    @patch("core.middleware.channels.logger")
    async def test_token_auth_middleware_with_user_id(self, mock_logger):
        async def mock_inner(scope, receive, send):
            pass

        middleware = TokenAuthMiddleware(mock_inner)

        scope = {"query_string": b"user_id=123", "user": None}

        with patch("core.middleware.channels.get_user") as mock_get_user:
            mock_user = MagicMock()
            mock_user.is_anonymous = False
            mock_user.username = "testuser"
            mock_get_user.return_value = mock_user

            await middleware(scope, AsyncMock(), AsyncMock())

            mock_get_user.assert_called_once_with(123)
            self.assertEqual(scope["user"], mock_user)
            mock_logger.debug.assert_any_call(
                "No token found, but user_id is present: 123"
            )
            mock_logger.debug.assert_any_call("User retrieved by ID: testuser")

    @patch("core.middleware.channels.logger")
    async def test_token_auth_middleware_access_token_priority(
        self, mock_logger
    ):
        async def mock_inner(scope, receive, send):
            pass

        middleware = TokenAuthMiddleware(mock_inner)

        scope = {
            "query_string": b"access_token=access_token_value&session_token=session_token_value",
            "user": None,
        }

        with patch("core.middleware.channels.authenticate_token") as mock_auth:
            mock_user = MagicMock()
            mock_user.is_anonymous = False
            mock_user.username = "testuser"
            mock_auth.return_value = mock_user

            await middleware(scope, AsyncMock(), AsyncMock())

            mock_auth.assert_called_once_with("access_token_value")

    @patch("core.middleware.channels.logger")
    async def test_token_auth_middleware_no_auth(self, mock_logger):
        async def mock_inner(scope, receive, send):
            pass

        middleware = TokenAuthMiddleware(mock_inner)

        scope = {"query_string": b"other_param=value", "user": None}

        await middleware(scope, AsyncMock(), AsyncMock())

        self.assertIsInstance(scope["user"], AnonymousUser)
        mock_logger.debug.assert_any_call(
            "TokenAuthMiddleware Token not found and no user_id"
        )

    @patch("core.middleware.channels.logger")
    async def test_token_auth_middleware_no_query_string(self, mock_logger):
        async def mock_inner(scope, receive, send):
            pass

        middleware = TokenAuthMiddleware(mock_inner)

        scope = {"user": None}

        await middleware(scope, AsyncMock(), AsyncMock())

        self.assertIsInstance(scope["user"], AnonymousUser)
        mock_logger.debug.assert_any_call("TokenAuthMiddleware called")
        mock_logger.debug.assert_any_call("Query params: {}")

    @patch("core.middleware.channels.logger")
    async def test_token_auth_middleware_anonymous_result(self, mock_logger):
        async def mock_inner(scope, receive, send):
            pass

        middleware = TokenAuthMiddleware(mock_inner)

        scope = {"query_string": b"access_token=invalid_token", "user": None}

        with patch("core.middleware.channels.authenticate_token") as mock_auth:
            mock_auth.return_value = AnonymousUser()

            await middleware(scope, AsyncMock(), AsyncMock())

            self.assertIsInstance(scope["user"], AnonymousUser)
            mock_logger.debug.assert_any_call(
                "User authenticated: AnonymousUser"
            )

    def test_token_auth_middleware_stack(self):
        mock_inner = MagicMock()
        result = TokenAuthMiddlewareStack(mock_inner)

        self.assertIsInstance(result, TokenAuthMiddleware)
        self.assertEqual(result.inner, mock_inner)

    @patch("core.middleware.channels.logger")
    async def test_authenticate_token_expired_cleanup(self, mock_logger):
        user = await self.create_test_user()
        token_string = "test_key123456789abcdef"
        token_key = token_string[:15]
        expired_time = timezone.now() - timedelta(hours=1)
        token = await self.create_test_token(
            user, token_key=token_key, expiry=expired_time
        )

        result = await authenticate_token(token_string)

        self.assertIsInstance(result, AnonymousUser)
        mock_logger.debug.assert_any_call("Token expired and was cleaned up")

        with self.assertRaises(Token.DoesNotExist):
            await database_sync_to_async(Token.objects.get)(pk=token.pk)

    @patch("core.middleware.channels.logger")
    async def test_token_auth_middleware_logging_flow(self, mock_logger):
        async def mock_inner(scope, receive, send):
            pass

        middleware = TokenAuthMiddleware(mock_inner)

        scope = {
            "query_string": b"access_token=test_token&session_token=session_token&user_id=123",
            "user": None,
        }

        with patch("core.middleware.channels.authenticate_token") as mock_auth:
            mock_user = MagicMock()
            mock_user.is_anonymous = False
            mock_user.username = "testuser"
            mock_auth.return_value = mock_user

            await middleware(scope, AsyncMock(), AsyncMock())

            mock_logger.debug.assert_any_call("TokenAuthMiddleware called")
            mock_logger.debug.assert_any_call(
                "Query params: {'access_token': ['test_token'], 'session_token': ['session_token'], 'user_id': ['123']}"
            )
            mock_logger.debug.assert_any_call("Session token: session_to...")
            mock_logger.debug.assert_any_call("Access token: test_token...")
            mock_logger.debug.assert_any_call("User ID: 123")
            mock_logger.debug.assert_any_call(
                "TokenAuthMiddleware Token found, authenticating"
            )
            mock_logger.debug.assert_any_call("User authenticated: testuser")



from drf_spectacular.utils import extend_schema_field
from parler_rest.serializers import TranslatableModelSerializer
from rest_framework import serializers
from rest_framework.relations import PrimaryKeyRelatedField

from core.api.schema import generate_schema_multi_lang
from core.utils.serializers import TranslatedFieldExtended
from product.models.category import ProductCategory
from product.models.category_image import ProductCategoryImage


@extend_schema_field(generate_schema_multi_lang(ProductCategoryImage))
class TranslatedFieldsFieldExtend(TranslatedFieldExtended):
    pass


class ProductCategoryImageSerializer(
    TranslatableModelSerializer,
    serializers.ModelSerializer[ProductCategoryImage],
):
    category = PrimaryKeyRelatedField(queryset=ProductCategory.objects.all())
    translations = TranslatedFieldsFieldExtend(
        shared_model=ProductCategoryImage
    )
    category_name = serializers.CharField(
        source="category.name", read_only=True
    )

    class Meta:
        model = ProductCategoryImage
        fields = (
            "id",
            "category",
            "category_name",
            "image",
            "image_type",
            "active",
            "sort_order",
            "translations",
            "image_path",
            "image_url",
            "created_at",
            "updated_at",
            "uuid",
        )
        read_only_fields = (
            "id",
            "category_name",
            "image_path",
            "image_url",
            "created_at",
            "updated_at",
            "uuid",
        )


class ProductCategoryImageDetailSerializer(ProductCategoryImageSerializer):
    class Meta(ProductCategoryImageSerializer.Meta):
        fields = (
            *ProductCategoryImageSerializer.Meta.fields,
            "translations",
        )


class ProductCategoryImageWriteSerializer(
    TranslatableModelSerializer,
    serializers.ModelSerializer[ProductCategoryImage],
):
    category = PrimaryKeyRelatedField(queryset=ProductCategory.objects.all())
    translations = TranslatedFieldsFieldExtend(
        shared_model=ProductCategoryImage
    )

    class Meta:
        model = ProductCategoryImage
        fields = (
            "category",
            "image",
            "image_type",
            "active",
            "sort_order",
            "translations",
        )

    def validate(self, attrs):
        category = attrs.get("category")
        image_type = attrs.get("image_type")

        if category and image_type:
            existing = ProductCategoryImage.objects.filter(
                category=category, image_type=image_type
            ).exclude(pk=self.instance.pk if self.instance else None)

            if existing.exists():
                raise serializers.ValidationError(
                    f"An image of type '{image_type}' already exists for this category."
                )

        return super().validate(attrs)


class ProductCategoryImageResponseSerializer(serializers.Serializer):
    success = serializers.BooleanField()
    message = serializers.CharField()
    data = ProductCategoryImageDetailSerializer()


class ProductCategoryImageBulkUpdateSerializer(serializers.Serializer):
    image_ids = serializers.ListField(
        child=serializers.IntegerField(), allow_empty=False
    )
    active = serializers.BooleanField(required=False)
    sort_order = serializers.IntegerField(required=False)

    def validate_image_ids(self, value: list[int]) -> list[int]:
        if not value:
            raise serializers.ValidationError(
                "At least one image ID is required."
            )

        existing_ids = ProductCategoryImage.objects.filter(
            id__in=value
        ).values_list("id", flat=True)

        missing_ids = set(value) - set(existing_ids)
        if missing_ids:
            raise serializers.ValidationError(
                f"Images with IDs {list(missing_ids)} do not exist."
            )

        return value


class ProductCategoryImageBulkResponseSerializer(serializers.Serializer):
    success = serializers.BooleanField()
    message = serializers.CharField()
    updated_count = serializers.IntegerField()



import logging
from abc import ABC, abstractmethod

from django.conf import settings
from django.core.mail import send_mail
from django.template.loader import render_to_string
from django.utils.translation import gettext_lazy as _

from order.models.order import Order

logger = logging.getLogger(__name__)


class NotifierInterface(ABC):
    @abstractmethod
    def send_order_confirmation(self, order: Order) -> bool:
        pass

    @abstractmethod
    def send_order_shipped(
        self, order: Order, tracking_number: str, carrier: str
    ) -> bool:
        pass

    @abstractmethod
    def send_order_delivered(self, order: Order) -> bool:
        pass

    @abstractmethod
    def send_order_canceled(self, order: Order) -> bool:
        pass


class EmailNotifier(NotifierInterface):
    def send_order_confirmation(self, order: Order) -> bool:
        subject = _("Order Confirmation #{order_id}").format(order_id=order.id)
        context = {
            "order": order,
            "items": order.items.all(),
        }
        html_message = render_to_string(
            "emails/order_confirmation.html", context
        )
        text_message = render_to_string(
            "emails/order_confirmation.txt", context
        )

        try:
            send_mail(
                subject=subject,
                message=text_message,
                html_message=html_message,
                from_email=settings.DEFAULT_FROM_EMAIL,
                recipient_list=[order.email],
                fail_silently=False,
            )
            return True
        except Exception as e:
            logger.error(f"Failed to send order confirmation email: {e!s}")
            return False

    def send_order_shipped(
        self, order: Order, tracking_number: str, carrier: str
    ) -> bool:
        subject = _("Your Order #{order_id} Has Shipped").format(
            order_id=order.id
        )
        context = {
            "order": order,
            "tracking_number": tracking_number,
            "carrier": carrier,
        }
        html_message = render_to_string("emails/order_shipped.html", context)
        text_message = render_to_string("emails/order_shipped.txt", context)

        try:
            send_mail(
                subject=subject,
                message=text_message,
                html_message=html_message,
                from_email=settings.DEFAULT_FROM_EMAIL,
                recipient_list=[order.email],
                fail_silently=False,
            )
            return True
        except Exception as e:
            logger.error(f"Failed to send order shipped email: {e!s}")
            return False

    def send_order_delivered(self, order: Order) -> bool:
        subject = _("Your Order #{order_id} Has Been Delivered").format(
            order_id=order.id
        )
        context = {
            "order": order,
        }
        html_message = render_to_string("emails/order_delivered.html", context)
        text_message = render_to_string("emails/order_delivered.txt", context)

        try:
            send_mail(
                subject=subject,
                message=text_message,
                html_message=html_message,
                from_email=settings.DEFAULT_FROM_EMAIL,
                recipient_list=[order.email],
                fail_silently=False,
            )
            return True
        except Exception as e:
            logger.error(f"Failed to send order delivered email: {e!s}")
            return False

    def send_order_canceled(self, order: Order) -> bool:
        subject = _("Your Order #{order_id} Has Been Canceled").format(
            order_id=order.id
        )
        context = {
            "order": order,
        }
        html_message = render_to_string("emails/order_canceled.html", context)
        text_message = render_to_string("emails/order_canceled.txt", context)

        try:
            send_mail(
                subject=subject,
                message=text_message,
                html_message=html_message,
                from_email=settings.DEFAULT_FROM_EMAIL,
                recipient_list=[order.email],
                fail_silently=False,
            )
            return True
        except Exception as e:
            logger.error(f"Failed to send order canceled email: {e!s}")
            return False


class SMSNotifier(NotifierInterface):
    def send_order_confirmation(self, order: Order) -> bool:
        if not order.phone:
            return False

        message = _(
            "Thank you for your order #{order_id}. Your order total is {total_price}. We'll notify you when your order ships."
        ).format(order_id=order.id, total_price=order.total_price)

        try:
            # @TODO - Call SMS provider API here
            print(message)
            return True
        except Exception as e:
            logger.error(f"Failed to send order confirmation SMS: {e!s}")
            return False

    def send_order_shipped(
        self, order: Order, tracking_number: str = "", carrier: str = ""
    ) -> bool:
        if not order.phone:
            return False

        message = _(
            "Your order #{order_id} has shipped! Track it with {carrier} using tracking number {tracking_number}."
        ).format(
            order_id=order.id, carrier=carrier, tracking_number=tracking_number
        )

        try:
            # @TODO - Call SMS provider API here
            print(message)
            return True
        except Exception as e:
            logger.error(f"Failed to send order shipped SMS: {e!s}")
            return False

    def send_order_delivered(self, order: Order) -> bool:
        if not order.phone:
            return False

        message = _(
            "Your order #{order_id} has been delivered! Thank you for shopping with us."
        ).format(order_id=order.id)

        try:
            # @TODO - Call SMS provider API here
            print(message)
            return True
        except Exception as e:
            logger.error(f"Failed to send order delivered SMS: {e!s}")
            return False

    def send_order_canceled(self, order: Order) -> bool:
        if not order.phone:
            return False

        message = _(
            "Your order #{order_id} has been canceled. Please contact customer service if you have any questions."
        ).format(order_id=order.id)

        try:
            # @TODO - Call SMS provider API here
            print(message)
            return True
        except Exception as e:
            logger.error(f"Failed to send order canceled SMS: {e!s}")
            return False


class OrderNotificationManager:
    def __init__(self):
        self.email_notifier = EmailNotifier()
        self.sms_notifier = SMSNotifier()

    def send_order_confirmation(self, order: Order) -> bool:
        email_sent = self.email_notifier.send_order_confirmation(order)

        sms_sent = False
        if order.phone:
            sms_sent = self.sms_notifier.send_order_confirmation(order)

        return email_sent or sms_sent

    def send_order_shipped(
        self, order: Order, tracking_number: str, carrier: str
    ) -> bool:
        email_sent = self.email_notifier.send_order_shipped(
            order, tracking_number, carrier
        )

        sms_sent = False
        if order.phone:
            sms_sent = self.sms_notifier.send_order_shipped(
                order, tracking_number=tracking_number, carrier=carrier
            )

        return email_sent or sms_sent

    def send_order_delivered(self, order: Order) -> bool:
        email_sent = self.email_notifier.send_order_delivered(order)

        sms_sent = False
        if order.phone:
            sms_sent = self.sms_notifier.send_order_delivered(order)

        return email_sent or sms_sent

    def send_order_canceled(self, order: Order) -> bool:
        email_sent = self.email_notifier.send_order_canceled(order)

        sms_sent = False
        if order.phone:
            sms_sent = self.sms_notifier.send_order_canceled(order)

        return email_sent or sms_sent


def send_order_confirmation(order: Order) -> bool:
    manager = OrderNotificationManager()
    return manager.send_order_confirmation(order)


def send_order_shipped_notification(order: Order) -> bool:
    manager = OrderNotificationManager()
    return manager.send_order_shipped(
        order,
        tracking_number=order.tracking_number,
        carrier=order.shipping_carrier,
    )


def send_order_delivered_notification(order: Order) -> bool:
    manager = OrderNotificationManager()
    return manager.send_order_delivered(order)


def send_order_canceled_notification(order: Order) -> bool:
    manager = OrderNotificationManager()
    return manager.send_order_canceled(order)



from django.urls import reverse
from rest_framework.request import Request
from rest_framework.test import APITestCase

from blog.factories import BlogAuthorFactory, BlogPostFactory
from blog.models.post import BlogPost
from blog.views.post import BlogPostViewSet
from core.filters.camel_case_ordering import CamelCaseOrderingFilter
from user.factories import UserAccountFactory


class CamelCaseOrderingFilterTest(APITestCase):
    def setUp(self):
        self.view = BlogPostViewSet.as_view({"get": "list"})
        self.filter = CamelCaseOrderingFilter()
        self.author1 = BlogAuthorFactory(user=UserAccountFactory())
        self.author2 = BlogAuthorFactory(user=UserAccountFactory())

        BlogPostFactory(author=self.author1)
        BlogPostFactory(author=self.author2)

    def _get_ordering(self, response, queryset):
        drf_request = Request(response.wsgi_request)
        view = BlogPostViewSet()
        view.request = drf_request
        view.format_kwarg = None
        view.kwargs = {}
        view.action = "list"
        return self.filter.get_ordering(drf_request, queryset, view)

    def _get_default_ordering(self):
        view = BlogPostViewSet()
        return self.filter.get_default_ordering(view)

    def create_request(self, ordering_value):
        url = reverse("blog-post-list")
        return self.client.get(url, {"ordering": ordering_value})

    def test_get_ordering_camel_case(self):
        request = self.create_request("createdAt")
        queryset = BlogPost.objects.all()
        result = self._get_ordering(request, queryset)
        self.assertEqual(result, ["created_at"])

    def test_get_ordering_pascal_case(self):
        request = self.create_request("CreatedAt")
        queryset = BlogPost.objects.all()
        result = self._get_ordering(request, queryset)
        self.assertEqual(result, ["created_at"])

    def test_get_ordering_mixed_case(self):
        request = self.create_request("CreatedAt,updatedAt")
        queryset = BlogPost.objects.all()
        result = self._get_ordering(request, queryset)
        self.assertEqual(result, ["created_at", "updated_at"])

    def test_get_ordering_snake_case(self):
        request = self.create_request("created_at,updated_at")
        queryset = BlogPost.objects.all()
        result = self._get_ordering(request, queryset)
        self.assertEqual(result, ["created_at", "updated_at"])

    def test_get_ordering_with_direction(self):
        request = self.create_request("-createdAt,updatedAt")
        queryset = BlogPost.objects.all()
        result = self._get_ordering(request, queryset)
        self.assertEqual(result, ["-created_at", "updated_at"])

    def test_get_ordering_mixed_directions(self):
        request = self.create_request("-CreatedAt,updatedAt,-viewCount")
        queryset = BlogPost.objects.all()
        result = self._get_ordering(request, queryset)
        self.assertEqual(result, ["-created_at", "updated_at", "-view_count"])

    def test_get_ordering_empty(self):
        request = self.create_request("")
        queryset = BlogPost.objects.all()
        result = self._get_ordering(request, queryset)
        default_ordering = self._get_default_ordering()
        self.assertEqual(result, default_ordering)
        self.assertEqual(result, ["-created_at"])

    def test_get_ordering_no_param(self):
        url = reverse("blog-post-list")
        request = self.client.get(url)
        queryset = BlogPost.objects.all()
        result = self._get_ordering(request, queryset)
        default_ordering = self._get_default_ordering()
        self.assertEqual(result, default_ordering)
        self.assertEqual(result, ["-created_at"])

    def test_get_ordering_invalid_field(self):
        request = self.create_request("createdAt,invalidField,-updatedAt")
        queryset = BlogPost.objects.all()
        result = self._get_ordering(request, queryset)
        self.assertEqual(result, ["created_at", "-updated_at"])

    def test_camel_to_snake_conversion(self):
        test_cases = [
            ("createdAt", "created_at"),
            ("CreatedAt", "created_at"),
            ("HTTPResponse", "http_response"),
            ("getHTTPResponseCode", "get_http_response_code"),
            ("already_snake_case", "already_snake_case"),
            ("alreadySnakeCase", "already_snake_case"),
            ("IOError", "io_error"),
            ("XMLParser", "xml_parser"),
            ("findXMLParser", "find_xml_parser"),
            ("id", "id"),
            ("ID", "id"),
            ("sortOrder", "sort_order"),
            ("isPublished", "is_published"),
            ("HTTPSProxy", "https_proxy"),
            ("HTTPAPI", "httpapi"),
            ("isHTTPSEnabled", "is_https_enabled"),
            ("viewCount", "view_count"),
            ("likesCount", "likes_count"),
            ("commentsCount", "comments_count"),
            ("tagsCount", "tags_count"),
            ("publishedAt", "published_at"),
        ]

        for input_str, expected in test_cases:
            with self.subTest(input=input_str):
                result = CamelCaseOrderingFilter.camel_to_snake(input_str)
                self.assertEqual(result, expected)

    def test_actual_api_ordering(self):
        url = reverse("blog-post-list")

        response = self.client.get(url, {"ordering": "createdAt"})
        self.assertEqual(response.status_code, 200)

        response = self.client.get(url, {"ordering": "CreatedAt"})
        self.assertEqual(response.status_code, 200)

        response = self.client.get(url, {"ordering": "-createdAt"})
        self.assertEqual(response.status_code, 200)

        response = self.client.get(url, {"ordering": "-publishedAt,viewCount"})
        self.assertEqual(response.status_code, 200)
        self.assertEqual(response.status_code, 200)



from .author import BlogAuthor
from .category import BlogCategory
from .comment import BlogComment
from .post import BlogPost
from .tag import BlogTag

__all__ = [
    "BlogAuthor",
    "BlogCategory",
    "BlogComment",
    "BlogPost",
    "BlogTag",
]



from __future__ import annotations

from datetime import timedelta
from functools import cached_property
from importlib import import_module
from typing import TYPE_CHECKING
from django.conf import settings
from django.db.models import F, Count, Q
from django.utils import timezone
from django.utils.translation import gettext_lazy as _
from drf_spectacular.utils import (
    OpenApiParameter,
    extend_schema,
    extend_schema_view,
)
from rest_framework import status
from rest_framework.decorators import action

from rest_framework.response import Response

from blog.filters.comment import BlogCommentFilter
from blog.filters.post import BlogPostFilter
from blog.models.post import BlogPost
from blog.serializers.comment import BlogCommentSerializer
from blog.serializers.post import (
    BlogPostDetailSerializer,
    BlogPostLikedPostsRequestSerializer,
    BlogPostLikedPostsResponseSerializer,
    BlogPostSerializer,
    BlogPostWriteSerializer,
)
from blog.strategies.weighted_related_posts_strategy import (
    WeightedRelatedPostsStrategy,
)
from core.api.serializers import ErrorResponseSerializer
from core.api.views import BaseModelViewSet

from core.utils.serializers import (
    create_schema_view_config,
    RequestSerializersConfig,
    ResponseSerializersConfig,
)
from core.utils.views import cache_methods

if TYPE_CHECKING:
    from blog.strategies.related_posts_strategy import RelatedPostsStrategy

req_serializers: RequestSerializersConfig = {
    "create": BlogPostWriteSerializer,
    "update": BlogPostWriteSerializer,
    "partial_update": BlogPostWriteSerializer,
    "liked_posts": BlogPostLikedPostsRequestSerializer,
    "update_likes": None,
    "update_view_count": None,
}

res_serializers: ResponseSerializersConfig = {
    "create": BlogPostDetailSerializer,
    "list": BlogPostSerializer,
    "retrieve": BlogPostDetailSerializer,
    "update": BlogPostDetailSerializer,
    "partial_update": BlogPostDetailSerializer,
    "update_likes": BlogPostDetailSerializer,
    "update_view_count": BlogPostDetailSerializer,
    "related_posts": BlogPostSerializer,
    "liked_posts": BlogPostLikedPostsResponseSerializer,
    "comments": BlogCommentSerializer,
    "trending": BlogPostSerializer,
    "popular": BlogPostSerializer,
    "featured": BlogPostSerializer,
}


@extend_schema_view(
    **create_schema_view_config(
        model_class=BlogPost,
        display_config={
            "tag": "Blog Post",
        },
        request_serializers=req_serializers,
        response_serializers=res_serializers,
        error_serializer=ErrorResponseSerializer,
    )
)
@cache_methods(settings.DEFAULT_CACHE_TTL, methods=["list", "retrieve"])
class BlogPostViewSet(BaseModelViewSet):
    queryset = BlogPost.objects.all()
    request_serializers = req_serializers
    response_serializers = res_serializers

    ordering_fields = [
        "id",
        "created_at",
        "updated_at",
        "published_at",
        "view_count",
        "featured",
    ]
    ordering = ["-created_at"]
    search_fields = [
        "translations__title",
        "translations__subtitle",
        "translations__body",
        "category__translations__name",
        "tags__translations__name",
    ]

    def get_queryset(self):
        queryset = BlogPost.objects.select_related(
            "author__user", "category", "author"
        ).prefetch_related(
            "tags__translations",
            "likes",
            "comments",
            "translations",
            "category__translations",
        )
        return queryset

    def get_filterset_class(self):
        if self.action == "comments":
            return None
        return BlogPostFilter

    @property
    def filterset_class(self):
        return self.get_filterset_class()

    @cached_property
    def related_posts_strategy(self):
        strategies_with_weights: list[tuple[RelatedPostsStrategy, float]] = []
        for strategy_config in settings.RELATED_POSTS_STRATEGIES:
            strategy_path = strategy_config["strategy"]
            weight = strategy_config["weight"]
            module_path, class_name = strategy_path.rsplit(".", 1)
            module = import_module(module_path)
            strategy_class = getattr(module, class_name)
            strategy_instance = strategy_class()
            strategies_with_weights.append((strategy_instance, weight))

        limit = getattr(settings, "RELATED_POSTS_LIMIT", 8)
        return WeightedRelatedPostsStrategy(
            strategies_with_weights, limit=limit
        )

    def get_related_posts_strategy(self):
        return self.related_posts_strategy

    @extend_schema(
        operation_id="toggleBlogPostLike",
        summary=_("Toggle post like"),
        description=_("Like or unlike a blog post. Toggles the like status."),
        tags=["Blog Posts"],
        request=None,
        responses={
            200: BlogPostDetailSerializer,
            401: ErrorResponseSerializer,
            404: ErrorResponseSerializer,
        },
    )
    @action(detail=True, methods=["POST"])
    def update_likes(self, request, pk=None):
        if not request.user.is_authenticated:
            return Response(
                {"detail": _("Authentication credentials were not provided.")},
                status=status.HTTP_401_UNAUTHORIZED,
            )

        post = self.get_object()
        user = request.user

        if post.likes.filter(pk=user.pk).exists():
            post.likes.remove(user)
        else:
            post.likes.add(user)
        post.save()

        response_serializer_class = self.get_response_serializer()
        response_serializer = response_serializer_class(
            post, context=self.get_serializer_context()
        )
        return Response(response_serializer.data, status=status.HTTP_200_OK)

    @extend_schema(
        operation_id="incrementBlogPostViews",
        summary=_("Increment post view count"),
        description=_("Increment the view count for a blog post."),
        tags=["Blog Posts"],
        responses={
            200: BlogPostDetailSerializer,
            404: ErrorResponseSerializer,
        },
    )
    @action(detail=True, methods=["POST"])
    def update_view_count(self, request, pk=None):
        post = self.get_object()
        post.view_count += 1
        post.save(update_fields=["view_count"])

        response_serializer_class = self.get_response_serializer()
        response_serializer = response_serializer_class(
            post, context=self.get_serializer_context()
        )
        return Response(response_serializer.data, status=status.HTTP_200_OK)

    @extend_schema(
        operation_id="listBlogPostRelated",
        summary=_("Get related posts"),
        description=_("Get related posts for a blog post."),
        tags=["Blog Posts"],
        parameters=[
            OpenApiParameter(
                name="id",
                description="A unique integer value identifying this blog post.",
                required=True,
                type=int,
                location=OpenApiParameter.PATH,
            ),
        ],
        responses={
            200: {
                "type": "array",
                "items": {"$ref": "#/components/schemas/BlogPost"},
            },
            404: ErrorResponseSerializer,
        },
    )
    @action(detail=True, methods=["GET"])
    def related_posts(self, request, pk=None):
        post = self.get_object()
        strategy = self.get_related_posts_strategy()
        related_posts = strategy.get_related_posts(post)

        response_serializer_class = self.get_response_serializer()
        response_serializer = response_serializer_class(
            related_posts, many=True, context=self.get_serializer_context()
        )
        return Response(response_serializer.data, status=status.HTTP_200_OK)

    @extend_schema(
        operation_id="checkBlogPostLikes",
        summary=_("Get liked posts"),
        description=_("Get all posts that the authenticated user has liked."),
        tags=["Blog Posts"],
        request=BlogPostLikedPostsRequestSerializer,
        responses={
            200: BlogPostLikedPostsResponseSerializer,
            400: ErrorResponseSerializer,
        },
    )
    @action(detail=False, methods=["POST"])
    def liked_posts(self, request, *args, **kwargs):
        request_serializer_class = self.get_request_serializer()
        request_serializer = request_serializer_class(data=request.data)
        if not request_serializer.is_valid():
            return Response(
                request_serializer.errors, status=status.HTTP_400_BAD_REQUEST
            )

        user = request.user
        post_ids = request_serializer.validated_data["post_ids"]

        liked_post_ids = BlogPost.objects.filter(
            likes=user, id__in=post_ids
        ).values_list("id", flat=True)

        response_data = {"post_ids": list(liked_post_ids)}

        response_serializer_class = self.get_response_serializer()
        response_serializer = response_serializer_class(response_data)
        return Response(response_serializer.data, status=status.HTTP_200_OK)

    @extend_schema(
        operation_id="listBlogPostComments",
        summary=_("Get post comments"),
        description=_("Get all comments for a blog post."),
        tags=["Blog Posts"],
        parameters=[
            OpenApiParameter(
                name="parent",
                type=str,
                description="Parent comment ID",
                required=False,
            ),
            OpenApiParameter(
                name="parent__isnull",
                type=bool,
                description="Filter comments with no parent",
                required=False,
            ),
        ],
        responses={
            200: BlogCommentSerializer(many=True),
            404: ErrorResponseSerializer,
        },
    )
    @action(detail=True, methods=["GET"])
    def comments(self, request, pk=None):
        post = self.get_object()

        self.ordering_fields = []
        self.ordering = []
        self.search_fields = []

        queryset = post.comments.select_related(
            "user", "parent"
        ).prefetch_related("translations", "likes")

        # Filter out unapproved comments for non-staff users
        if not request.user.is_staff:
            queryset = queryset.filter(approved=True)

        comment_filterset = BlogCommentFilter(
            data=request.GET, queryset=queryset, request=request
        )

        if comment_filterset.is_valid():
            queryset = comment_filterset.qs
        else:
            return Response(
                {"filter_errors": comment_filterset.errors},
                status=status.HTTP_400_BAD_REQUEST,
            )

        response_serializer_class = self.get_response_serializer()
        return self.paginate_and_serialize(
            queryset, request, serializer_class=response_serializer_class
        )

    @extend_schema(
        operation_id="listTrendingBlogPosts",
        summary=_("Get trending posts"),
        description=_(
            "Get trending blog posts based on recent engagement metrics. "
            "Combines views, likes, and comments from recent time period."
        ),
        tags=["Blog Posts"],
        parameters=[
            OpenApiParameter(
                name="days",
                type=int,
                location=OpenApiParameter.QUERY,
                description="Number of days to look back for trending calculation",
                required=False,
                default=7,
            ),
        ],
        responses={
            200: BlogPostSerializer(many=True),
        },
    )
    @action(detail=False, methods=["GET"])
    def trending(self, request):
        days = int(request.query_params.get("days", 7))
        cutoff_date = timezone.now() - timedelta(days=days)

        queryset = (
            self.get_queryset()
            .filter(published_at__gte=cutoff_date)
            .annotate(
                likes_count_annotation=Count("likes", distinct=True),
                comments_count_annotation=Count(
                    "comments", distinct=True, filter=Q(comments__approved=True)
                ),
                trending_score=F("view_count")
                + (F("likes_count_annotation") * 2)
                + (F("comments_count_annotation") * 3),
            )
            .order_by("-trending_score")
        )

        response_serializer_class = self.get_response_serializer()
        return self.paginate_and_serialize(
            queryset, request, serializer_class=response_serializer_class
        )

    @extend_schema(
        operation_id="listPopularBlogPosts",
        summary=_("Get popular posts"),
        description=_(
            "Get most popular blog posts based on all-time engagement metrics."
        ),
        tags=["Blog Posts"],
        responses={
            200: BlogPostSerializer(many=True),
        },
    )
    @action(detail=False, methods=["GET"])
    def popular(self, request):
        queryset = (
            self.get_queryset()
            .annotate(likes_count_annotation=Count("likes", distinct=True))
            .order_by("-likes_count_annotation", "-view_count")
        )

        response_serializer_class = self.get_response_serializer()
        return self.paginate_and_serialize(
            queryset, request, serializer_class=response_serializer_class
        )

    @extend_schema(
        operation_id="listFeaturedBlogPosts",
        summary=_("Get featured posts"),
        description=_(
            "Get posts marked as featured, ordered by publication date."
        ),
        tags=["Blog Posts"],
        responses={
            200: BlogPostSerializer(many=True),
        },
    )
    @action(detail=False, methods=["GET"])
    def featured(self, request):
        queryset = (
            self.get_queryset().filter(featured=True).order_by("-published_at")
        )

        response_serializer_class = self.get_response_serializer()
        return self.paginate_and_serialize(
            queryset, request, serializer_class=response_serializer_class
        )



# Generated by Django 4.2.9 on 2024-02-09 23:38
from django.db import migrations
from django.db import models


class Migration(migrations.Migration):
    initial = True

    dependencies = [
        ("core", "0001_pg_trigram_extension"),
    ]

    operations = [
        migrations.CreateModel(
            name="Settings",
            fields=[
                (
                    "id",
                    models.BigAutoField(
                        auto_created=True,
                        primary_key=True,
                        serialize=False,
                        verbose_name="ID",
                    ),
                ),
                (
                    "key",
                    models.CharField(max_length=255, unique=True, verbose_name="Key"),
                ),
                ("value", models.TextField(verbose_name="Value")),
                (
                    "value_type",
                    models.CharField(
                        choices=[
                            ("str", "String"),
                            ("int", "Integer"),
                            ("bool", "Boolean"),
                            ("dict", "Dictionary"),
                            ("list", "List"),
                            ("float", "Float"),
                        ],
                        default="str",
                        max_length=50,
                        verbose_name="Value Type",
                    ),
                ),
                (
                    "description",
                    models.TextField(blank=True, null=True, verbose_name="Description"),
                ),
                (
                    "is_public",
                    models.BooleanField(default=True, verbose_name="Is Public"),
                ),
            ],
            options={
                "verbose_name": "Setting",
                "verbose_name_plural": "Settings",
            },
        ),
    ]



# Generated by Django 5.0.3 on 2024-04-01 12:25
from django.db import migrations


class Migration(migrations.Migration):
    dependencies = [
        ("blog", "0006_alter_blogpost_view_count"),
    ]

    operations = [
        migrations.RemoveConstraint(
            model_name="blogcomment",
            name="unique_blog_comment",
        ),
    ]



from django.test import TestCase
from mptt.managers import TreeManager
from parler.managers import TranslatableManager

from blog.factories.author import BlogAuthorFactory
from blog.factories.comment import BlogCommentFactory
from blog.factories.post import BlogPostFactory
from blog.managers.comment import BlogCommentManager, BlogCommentQuerySet
from blog.models.comment import BlogComment
from user.factories.account import UserAccountFactory


class TestBlogCommentManager(TestCase):
    def setUp(self):
        self.author = BlogAuthorFactory(user=UserAccountFactory())
        self.post = BlogPostFactory(author=self.author)

        self.user1 = UserAccountFactory()
        self.user2 = UserAccountFactory()
        self.user3 = UserAccountFactory()
        self.user4 = UserAccountFactory()
        self.user5 = UserAccountFactory()

        self.approved_comment1 = BlogCommentFactory(
            post=self.post, user=self.user1, approved=True
        )
        self.approved_comment2 = BlogCommentFactory(
            post=self.post, user=self.user2, approved=True
        )
        self.unapproved_comment = BlogCommentFactory(
            post=self.post, user=self.user3, approved=False
        )

        self.reply_approved = BlogCommentFactory(
            post=self.post,
            user=self.user4,
            parent=self.approved_comment1,
            approved=True,
        )
        self.reply_unapproved = BlogCommentFactory(
            post=self.post,
            user=self.user5,
            parent=self.approved_comment1,
            approved=False,
        )

    def test_approved_manager_method(self):
        approved_comments = BlogComment.objects.approved()

        assert self.approved_comment1 in approved_comments
        assert self.approved_comment2 in approved_comments
        assert self.reply_approved in approved_comments
        assert self.unapproved_comment not in approved_comments
        assert self.reply_unapproved not in approved_comments

        assert approved_comments.count() == 3

    def test_manager_returns_correct_queryset_type(self):
        queryset = BlogComment.objects.approved()

        assert isinstance(queryset, BlogCommentQuerySet)

    def test_manager_inheritance(self):
        manager = BlogComment.objects
        assert isinstance(manager, BlogCommentManager)
        assert isinstance(manager, TreeManager)
        assert isinstance(manager, TranslatableManager)


class TestBlogCommentQuerySet(TestCase):
    def setUp(self):
        self.author = BlogAuthorFactory(user=UserAccountFactory())
        self.post = BlogPostFactory(author=self.author)

        self.user1 = UserAccountFactory()
        self.user2 = UserAccountFactory()
        self.user3 = UserAccountFactory()
        self.user4 = UserAccountFactory()
        self.user5 = UserAccountFactory()

        self.approved_comment1 = BlogCommentFactory(
            post=self.post, user=self.user1, approved=True
        )
        self.approved_comment2 = BlogCommentFactory(
            post=self.post, user=self.user2, approved=True
        )
        self.unapproved_comment = BlogCommentFactory(
            post=self.post, user=self.user3, approved=False
        )

        self.reply_approved = BlogCommentFactory(
            post=self.post,
            user=self.user4,
            parent=self.approved_comment1,
            approved=True,
        )
        self.reply_unapproved = BlogCommentFactory(
            post=self.post,
            user=self.user5,
            parent=self.approved_comment1,
            approved=False,
        )

    def test_approved_queryset_method(self):
        queryset = BlogComment.objects.filter().approved()

        assert self.approved_comment1 in queryset
        assert self.approved_comment2 in queryset
        assert self.reply_approved in queryset
        assert self.unapproved_comment not in queryset
        assert self.reply_unapproved not in queryset

        assert queryset.count() == 3

    def test_queryset_chaining_with_approved(self):
        queryset = BlogComment.objects.filter(user=self.user1).approved()

        assert self.approved_comment1 in queryset
        assert self.unapproved_comment not in queryset

        assert queryset.count() >= 1

    def test_queryset_with_tree_methods(self):
        queryset = BlogComment.objects.approved()

        root_comments = queryset.filter(parent__isnull=True)
        assert root_comments.count() == 2

        descendants = self.approved_comment1.get_descendants(include_self=True)
        approved_descendants = queryset.filter(
            id__in=descendants.values_list("id", flat=True)
        )
        assert approved_descendants.count() == 2

    def test_queryset_with_translations(self):
        queryset = (
            BlogComment.objects.approved()
            .exclude(translations__content__isnull=True)
            .exclude(translations__content__exact="")
        )

        assert self.approved_comment1 in queryset
        assert self.approved_comment2 in queryset
        assert self.reply_approved in queryset
        assert self.unapproved_comment not in queryset

    def test_as_manager_class_method(self):
        manager = BlogCommentQuerySet.as_manager()
        assert hasattr(manager, "_built_with_as_manager")
        assert manager._built_with_as_manager is True

    def test_queryset_returns_correct_type(self):
        queryset = BlogComment.objects.filter().approved()

        assert isinstance(queryset, BlogCommentQuerySet)

        chained = queryset.filter(user=self.user1)
        assert isinstance(chained, BlogCommentQuerySet)


class TestBlogCommentManagerTreeFunctionality(TestCase):
    def setUp(self):
        self.author = BlogAuthorFactory(user=UserAccountFactory())
        self.post = BlogPostFactory(author=self.author)

        self.user1 = UserAccountFactory()
        self.user2 = UserAccountFactory()
        self.user3 = UserAccountFactory()
        self.user4 = UserAccountFactory()
        self.user5 = UserAccountFactory()

        self.root1 = BlogCommentFactory(
            post=self.post, user=self.user1, approved=True
        )
        self.root2 = BlogCommentFactory(
            post=self.post, user=self.user2, approved=False
        )

        self.child1_1 = BlogCommentFactory(
            post=self.post, user=self.user3, parent=self.root1, approved=True
        )
        self.child1_2 = BlogCommentFactory(
            post=self.post,
            user=self.user4,
            parent=self.root1,
            approved=False,
        )

        self.grandchild1_1_1 = BlogCommentFactory(
            post=self.post,
            user=self.user5,
            parent=self.child1_1,
            approved=True,
        )

    def test_approved_with_tree_structure(self):
        approved_comments = BlogComment.objects.approved()

        assert self.root1 in approved_comments
        assert self.child1_1 in approved_comments
        assert self.grandchild1_1_1 in approved_comments

        assert self.root2 not in approved_comments
        assert self.child1_2 not in approved_comments

        assert approved_comments.count() == 3

    def test_tree_methods_with_approved_filter(self):
        all_descendants = self.root1.get_descendants(include_self=True)
        approved_descendants = BlogComment.objects.approved().filter(
            id__in=all_descendants.values_list("id", flat=True)
        )

        assert approved_descendants.count() == 3
        assert self.root1 in approved_descendants
        assert self.child1_1 in approved_descendants
        assert self.grandchild1_1_1 in approved_descendants
        assert self.child1_2 not in approved_descendants

    def test_tree_level_filtering_with_approved(self):
        root_approved = BlogComment.objects.approved().filter(level=0)
        assert root_approved.count() == 1
        assert self.root1 in root_approved
        assert self.root2 not in root_approved

        level1_approved = BlogComment.objects.approved().filter(level=1)
        assert level1_approved.count() == 1
        assert self.child1_1 in level1_approved
        assert self.child1_2 not in level1_approved

        level2_approved = BlogComment.objects.approved().filter(level=2)
        assert level2_approved.count() == 1
        assert self.grandchild1_1_1 in level2_approved


class TestBlogCommentManagerEdgeCases(TestCase):
    def test_empty_queryset_behavior(self):
        assert BlogComment.objects.approved().count() == 0

        queryset = BlogComment.objects.approved()
        assert list(queryset) == []

    def test_all_unapproved_comments(self):
        author = BlogAuthorFactory(user=UserAccountFactory())
        post = BlogPostFactory(author=author)

        BlogCommentFactory(post=post, user=UserAccountFactory(), approved=False)
        BlogCommentFactory(post=post, user=UserAccountFactory(), approved=False)

        approved_comments = BlogComment.objects.approved()
        assert approved_comments.count() == 0

    def test_all_approved_comments(self):
        author = BlogAuthorFactory(user=UserAccountFactory())
        post = BlogPostFactory(author=author)

        comment1 = BlogCommentFactory(
            post=post, user=UserAccountFactory(), approved=True
        )
        comment2 = BlogCommentFactory(
            post=post, user=UserAccountFactory(), approved=True
        )
        comment3 = BlogCommentFactory(
            post=post, user=UserAccountFactory(), approved=True
        )

        approved_comments = BlogComment.objects.approved()
        assert approved_comments.count() == 3
        assert comment1 in approved_comments
        assert comment2 in approved_comments
        assert comment3 in approved_comments

    def test_manager_with_deleted_posts(self):
        author = BlogAuthorFactory(user=UserAccountFactory())
        post = BlogPostFactory(author=author)
        comment = BlogCommentFactory(
            post=post, user=UserAccountFactory(), approved=True
        )

        assert BlogComment.objects.approved().count() == 1

        post.delete()

        comment.refresh_from_db()
        assert comment.post is None

        approved_comments = BlogComment.objects.approved()
        assert approved_comments.count() == 1

    def test_manager_with_deleted_users(self):
        author = BlogAuthorFactory(user=UserAccountFactory())
        post = BlogPostFactory(author=author)
        user = UserAccountFactory()
        comment = BlogCommentFactory(post=post, user=user, approved=True)

        assert BlogComment.objects.approved().count() == 1

        user.delete()

        comment.refresh_from_db()
        assert comment.user is None

        approved_comments = BlogComment.objects.approved()
        assert approved_comments.count() == 1

    def test_queryset_class_attribute(self):
        manager = BlogComment.objects
        assert isinstance(manager, BlogCommentManager)
        assert manager._queryset_class == BlogCommentQuerySet

    def test_performance_with_large_tree(self):
        author = BlogAuthorFactory(user=UserAccountFactory())
        post = BlogPostFactory(author=author)

        root_comments = []
        for i in range(5):
            root = BlogCommentFactory(
                post=post,
                user=UserAccountFactory(),
                approved=i % 2 == 0,
            )
            root_comments.append(root)

            for j in range(3):
                child = BlogCommentFactory(
                    post=post,
                    user=UserAccountFactory(),
                    parent=root,
                    approved=j % 2 == 0,
                )

                for k in range(2):
                    BlogCommentFactory(
                        post=post,
                        user=UserAccountFactory(),
                        parent=child,
                        approved=k % 2 == 0,
                    )

        approved_comments = BlogComment.objects.approved()

        assert approved_comments.count() > 0

        for comment in approved_comments:
            assert comment.approved is True



from __future__ import annotations

import uuid
from enum import Enum, unique
from typing import TYPE_CHECKING

from cart.models import Cart, CartItem

if TYPE_CHECKING:
    from django.http import HttpRequest
    from rest_framework.request import Request

    from product.models.product import Product


class CartServiceInitException(Exception):
    pass


class CartNotSetException(Exception):
    pass


class InvalidProcessCartOptionException(Exception):
    pass


@unique
class ProcessCartOption(Enum):
    MERGE = "merge"
    CLEAN = "clean"
    KEEP = "keep"


class CartService:
    def __init__(self, request: Request | HttpRequest):
        if not request:
            raise CartServiceInitException(
                "Request must be provided to initialize CartService."
            )

        self.request = request
        self.cart: Cart | None = None
        self.cart_items: list[CartItem] = []

        self._extract_cart_info()

        try:
            self._initialize_cart()
        except CartServiceInitException:
            raise

    def _extract_cart_info(self):
        if hasattr(self.request, "META"):
            self.cart_id = self.request.META.get("HTTP_X_CART_ID")
            self.session_key = self.request.META.get("HTTP_X_SESSION_KEY")
        else:
            self.cart_id = self.request.headers.get("X-Cart-Id")
            self.session_key = self.request.headers.get("X-Session-Key")

        if self.cart_id:
            try:
                self.cart_id = int(self.cart_id)
            except (ValueError, TypeError):
                self.cart_id = None

    def __str__(self):
        return f"Cart {self.cart.user if self.cart and self.cart.user else 'Anonymous'}"

    def __len__(self):
        return self.cart.total_items if self.cart else 0

    def __iter__(self):
        if self.cart:
            yield from self.cart_items

    def _initialize_cart(self):
        self.cart = self.get_or_create_cart()
        self.cart_items = self.cart.get_items() if self.cart else []
        if self.cart:
            self.cart.refresh_last_activity()

    def get_or_create_cart(self):
        user = self.request.user

        if user.is_authenticated:
            cart, created = Cart.objects.get_or_create(user=user)

            if self.cart_id:
                guest_cart = Cart.objects.filter(
                    id=self.cart_id, user__isnull=True
                ).first()

                if guest_cart:
                    self.merge_carts(guest_cart, cart)

            return cart
        else:
            cart = None

            if self.cart_id:
                cart = Cart.objects.filter(
                    id=self.cart_id, user__isnull=True
                ).first()

                if (
                    cart
                    and self.session_key
                    and cart.session_key != self.session_key
                ):
                    cart = None

            if not cart and self.session_key:
                cart = Cart.objects.filter(
                    session_key=self.session_key, user__isnull=True
                ).first()

            if not cart:
                session_key = self.session_key or str(uuid.uuid4())
                cart = Cart.objects.create(session_key=session_key)

            return cart

    def merge_carts(self, source_cart: Cart, target_cart: Cart = None):
        if not target_cart:
            target_cart = self.cart

        if not target_cart or not source_cart:
            return

        if target_cart.id == source_cart.id:
            return

        for item in source_cart.items.all():
            existing_item = CartItem.objects.filter(
                cart=target_cart, product=item.product
            ).first()

            if existing_item:
                existing_item.quantity += item.quantity
                existing_item.save()
                item.delete()
            else:
                item.cart = target_cart
                item.save()

        if target_cart == self.cart:
            self.cart_items = self.cart.get_items()

        source_cart.delete()

    def clean_cart(self):
        if self.cart:
            self.cart.items.all().delete()
            self.cart_items = []

    def get_cart_by_id(self, cart_id: int):
        return Cart.objects.filter(id=cart_id).first()

    def get_cart_item(self, product_id: int | None):
        if self.cart:
            return self.cart.items.filter(product_id=product_id).first()
        return None

    def create_cart_item(self, product: Product, quantity: int):
        if not self.cart:
            raise CartNotSetException("Cart is not set.")

        existing_item = CartItem.objects.filter(
            cart=self.cart, product=product
        ).first()

        if existing_item:
            existing_item.quantity += quantity
            existing_item.save()
            cart_item = existing_item
        else:
            cart_item = CartItem.objects.create(
                cart=self.cart, product=product, quantity=quantity
            )

        self.cart_items = self.cart.get_items()
        return cart_item

    def update_cart_item(self, product_id: int, quantity: int):
        if not self.cart:
            raise CartNotSetException("Cart is not set.")
        cart_item = self.cart.items.get(product_id=product_id)
        cart_item.quantity = quantity
        cart_item.save()
        self.cart_items = self.cart.get_items()
        return cart_item

    def delete_cart_item(self, product_id: int):
        if not self.cart:
            raise CartNotSetException("Cart is not set.")
        self.cart.items.filter(product_id=product_id).delete()
        self.cart_items = self.cart.get_items()



from datetime import timedelta
from django.db.models import Q, F
from django.utils import timezone
from django.utils.translation import gettext_lazy as _
from django_filters import rest_framework as filters

from cart.models import CartItem
from core.filters.camel_case_filters import CamelCaseTimeStampFilterSet
from core.filters.core import UUIDFilterMixin


class CartItemFilter(UUIDFilterMixin, CamelCaseTimeStampFilterSet):
    cart = filters.NumberFilter(
        field_name="cart__id",
        lookup_expr="exact",
        help_text=_("Filter by cart ID"),
    )
    cart__uuid = filters.UUIDFilter(
        field_name="cart__uuid",
        help_text=_("Filter by cart UUID"),
    )
    cart__session_key = filters.CharFilter(
        field_name="cart__session_key",
        lookup_expr="exact",
        help_text=_("Filter by cart session key"),
    )
    cart__user = filters.NumberFilter(
        field_name="cart__user__id",
        lookup_expr="exact",
        help_text=_("Filter by cart user ID"),
    )
    cart__user__email = filters.CharFilter(
        field_name="cart__user__email",
        lookup_expr="icontains",
        help_text=_("Filter by cart user email (partial match)"),
    )
    cart__user__name = filters.CharFilter(
        method="filter_cart_user_name",
        help_text=_("Filter by cart user name (first or last)"),
    )
    cart__is_guest = filters.BooleanFilter(
        method="filter_cart_is_guest",
        help_text=_("Filter items in guest carts"),
    )

    product = filters.NumberFilter(
        field_name="product__id",
        lookup_expr="exact",
        help_text=_("Filter by product ID"),
    )
    product__uuid = filters.UUIDFilter(
        field_name="product__uuid",
        help_text=_("Filter by product UUID"),
    )
    product__sku = filters.CharFilter(
        field_name="product__sku",
        lookup_expr="iexact",
        help_text=_("Filter by product sku (exact match)"),
    )
    product__name = filters.CharFilter(
        field_name="product__translations__name",
        lookup_expr="icontains",
        help_text=_("Filter by product name (partial match)"),
    )
    product__category = filters.NumberFilter(
        field_name="product__category__id",
        help_text=_("Filter by product category ID"),
    )
    product__category__slug = filters.CharFilter(
        field_name="product__category__slug",
        help_text=_("Filter by product category slug"),
    )
    product__active = filters.BooleanFilter(
        field_name="product__active",
        help_text=_("Filter by product active status"),
    )

    quantity = filters.NumberFilter(
        field_name="quantity",
        help_text=_("Filter by exact quantity"),
    )
    min_quantity = filters.NumberFilter(
        field_name="quantity",
        lookup_expr="gte",
        help_text=_("Filter by minimum quantity"),
    )
    max_quantity = filters.NumberFilter(
        field_name="quantity",
        lookup_expr="lte",
        help_text=_("Filter by maximum quantity"),
    )

    min_price = filters.NumberFilter(
        field_name="product__price",
        lookup_expr="gte",
        help_text=_("Filter by minimum product price"),
    )
    max_price = filters.NumberFilter(
        field_name="product__price",
        lookup_expr="lte",
        help_text=_("Filter by maximum product price"),
    )
    min_total_price = filters.NumberFilter(
        method="filter_min_total_price",
        help_text=_("Filter by minimum total price (quantity * price)"),
    )
    max_total_price = filters.NumberFilter(
        method="filter_max_total_price",
        help_text=_("Filter by maximum total price (quantity * price)"),
    )

    with_discounts = filters.BooleanFilter(
        method="filter_with_discounts",
        help_text=_("Filter items with product discounts"),
    )
    min_discount_percent = filters.NumberFilter(
        field_name="product__discount_percent",
        lookup_expr="gte",
        help_text=_("Filter by minimum discount percentage"),
    )
    max_discount_percent = filters.NumberFilter(
        field_name="product__discount_percent",
        lookup_expr="lte",
        help_text=_("Filter by maximum discount percentage"),
    )

    in_active_carts = filters.BooleanFilter(
        method="filter_active_carts",
        help_text=_("Filter items in active carts (24hr)"),
    )
    in_abandoned_carts = filters.BooleanFilter(
        method="filter_abandoned_carts",
        help_text=_("Filter items in abandoned carts (30+ days)"),
    )
    cart_last_activity_after = filters.DateTimeFilter(
        field_name="cart__last_activity",
        lookup_expr="gte",
        help_text=_("Filter by cart last activity after date"),
    )
    cart_last_activity_before = filters.DateTimeFilter(
        field_name="cart__last_activity",
        lookup_expr="lte",
        help_text=_("Filter by cart last activity before date"),
    )

    class Meta:
        model = CartItem
        fields = {
            "created_at": ["gte", "lte", "date"],
            "updated_at": ["gte", "lte", "date"],
            "uuid": ["exact"],
            "id": ["exact", "in"],
            "quantity": ["exact", "gte", "lte"],
            "cart": ["exact"],
            "cart__user": ["exact"],
            "cart__session_key": ["exact", "icontains"],
            "product": ["exact"],
            "product__active": ["exact"],
        }

    def filter_cart_user_name(self, queryset, name, value):
        """Filter by cart user's first or last name."""
        if value:
            return queryset.filter(
                Q(cart__user__first_name__icontains=value)
                | Q(cart__user__last_name__icontains=value)
            )
        return queryset

    def filter_cart_is_guest(self, queryset, name, value):
        """Filter items in guest (no user) carts."""
        if value is True:
            return queryset.filter(cart__user__isnull=True)
        elif value is False:
            return queryset.filter(cart__user__isnull=False)
        return queryset

    def filter_min_total_price(self, queryset, name, value):
        """Filter by minimum total price (quantity * price)."""
        if value is not None:
            return queryset.annotate(
                total_price=F("quantity") * F("product__price")
            ).filter(total_price__gte=value)
        return queryset

    def filter_max_total_price(self, queryset, name, value):
        """Filter by maximum total price (quantity * price)."""
        if value is not None:
            return queryset.annotate(
                total_price=F("quantity") * F("product__price")
            ).filter(total_price__lte=value)
        return queryset

    def filter_with_discounts(self, queryset, name, value):
        """Filter items with product discounts."""
        if value is True:
            return queryset.filter(product__discount_percent__gt=0)
        elif value is False:
            return queryset.filter(
                Q(product__discount_percent=0)
                | Q(product__discount_percent__isnull=True)
            )
        return queryset

    def filter_active_carts(self, queryset, name, value):
        """Filter items in active carts (activity within 24 hours)."""
        cutoff_time = timezone.now() - timedelta(hours=24)
        if value is True:
            return queryset.filter(cart__last_activity__gte=cutoff_time)
        elif value is False:
            return queryset.filter(cart__last_activity__lt=cutoff_time)
        return queryset

    def filter_abandoned_carts(self, queryset, name, value):
        """Filter items in abandoned carts (inactive 30+ days)."""
        cutoff_time = timezone.now() - timedelta(days=30)
        if value is True:
            return queryset.filter(cart__last_activity__lt=cutoff_time)
        elif value is False:
            return queryset.filter(cart__last_activity__gte=cutoff_time)
        return queryset



from datetime import timedelta
from django.urls import reverse
from django.utils import timezone
from rest_framework.test import APITestCase

from user.factories.subscription import (
    SubscriptionTopicFactory,
    UserSubscriptionFactory,
)
from user.factories.account import UserAccountFactory
from user.models.subscription import SubscriptionTopic, UserSubscription


class SubscriptionTopicFilterTest(APITestCase):
    def setUp(self):
        UserSubscription.objects.all().delete()
        SubscriptionTopic.objects.all().delete()

        self.now = timezone.now()

        self.newsletter_topic = SubscriptionTopicFactory(
            slug="weekly-newsletter",
            category=SubscriptionTopic.TopicCategory.NEWSLETTER,
            is_active=True,
            is_default=True,
            requires_confirmation=False,
            set_translations=False,
        )
        self.newsletter_topic.created_at = self.now - timedelta(days=30)
        self.newsletter_topic.save()

        self.newsletter_topic.set_current_language("en")
        self.newsletter_topic.name = "Weekly Newsletter"
        self.newsletter_topic.description = "Get our weekly updates and news"
        self.newsletter_topic.save()

        self.marketing_topic = SubscriptionTopicFactory(
            slug="marketing-campaigns",
            category=SubscriptionTopic.TopicCategory.MARKETING,
            is_active=True,
            is_default=False,
            requires_confirmation=True,
            set_translations=False,
        )
        self.marketing_topic.created_at = self.now - timedelta(days=15)
        self.marketing_topic.save()

        self.marketing_topic.set_current_language("en")
        self.marketing_topic.name = "Marketing Updates"
        self.marketing_topic.description = (
            "Special offers and promotional content"
        )
        self.marketing_topic.save()

        self.product_topic = SubscriptionTopicFactory(
            slug="product-updates",
            category=SubscriptionTopic.TopicCategory.PRODUCT,
            is_active=False,
            is_default=False,
            requires_confirmation=False,
            set_translations=False,
        )
        self.product_topic.created_at = self.now - timedelta(days=5)
        self.product_topic.save()

        self.product_topic.set_current_language("en")
        self.product_topic.name = "Product News"
        self.product_topic.description = "Latest product features and updates"
        self.product_topic.save()

        self.system_topic = SubscriptionTopicFactory(
            slug="system-notifications",
            category=SubscriptionTopic.TopicCategory.SYSTEM,
            is_active=True,
            is_default=True,
            requires_confirmation=False,
            set_translations=False,
        )
        self.system_topic.created_at = self.now - timedelta(hours=2)
        self.system_topic.save()

        self.system_topic.set_current_language("en")
        self.system_topic.name = "System Alerts"
        self.system_topic.description = (
            "Important system notifications and maintenance updates"
        )
        self.system_topic.save()

        self.user1 = UserAccountFactory()
        self.user2 = UserAccountFactory()

        self.client.force_authenticate(user=self.user1)

    def test_timestamp_filters(self):
        url = reverse("user-subscription-topic-list")

        created_after_date = self.now - timedelta(days=20)
        response = self.client.get(
            url,
            {"created_after": created_after_date.isoformat()},
        )
        self.assertEqual(response.status_code, 200)

        result_ids = [r["id"] for r in response.data["results"]]
        self.assertNotIn(self.newsletter_topic.id, result_ids)
        self.assertIn(self.marketing_topic.id, result_ids)
        self.assertIn(self.system_topic.id, result_ids)
        self.assertGreaterEqual(len(result_ids), 2)

        created_before_date = self.now - timedelta(days=20)
        response = self.client.get(
            url,
            {"created_before": created_before_date.isoformat()},
        )
        self.assertEqual(response.status_code, 200)

        result_ids = [r["id"] for r in response.data["results"]]
        self.assertIn(self.newsletter_topic.id, result_ids)
        self.assertNotIn(self.marketing_topic.id, result_ids)
        self.assertNotIn(self.system_topic.id, result_ids)

    def test_uuid_filter(self):
        url = reverse("user-subscription-topic-list")

        response = self.client.get(
            url, {"uuid": str(self.newsletter_topic.uuid)}
        )
        self.assertEqual(response.status_code, 200)
        self.assertEqual(len(response.data["results"]), 1)
        self.assertEqual(
            response.data["results"][0]["id"], self.newsletter_topic.id
        )

    def test_camel_case_filters(self):
        url = reverse("user-subscription-topic-list")

        created_after_date = self.now - timedelta(days=20)
        response = self.client.get(
            url,
            {
                "createdAfter": created_after_date.isoformat(),
                "isActive": "true",
            },
        )
        self.assertEqual(response.status_code, 200)

        result_ids = [r["id"] for r in response.data["results"]]
        self.assertEqual(len(result_ids), 2)
        self.assertIn(self.marketing_topic.id, result_ids)
        self.assertIn(self.system_topic.id, result_ids)

        response = self.client.get(
            url,
            {
                "isDefault": "true",
                "requiresConfirmation": "false",
            },
        )
        self.assertEqual(response.status_code, 200)

        result_ids = [r["id"] for r in response.data["results"]]
        self.assertEqual(len(result_ids), 2)
        self.assertIn(self.newsletter_topic.id, result_ids)
        self.assertIn(self.system_topic.id, result_ids)

    def test_category_filters(self):
        url = reverse("user-subscription-topic-list")

        response = self.client.get(
            url, {"category": SubscriptionTopic.TopicCategory.NEWSLETTER}
        )
        self.assertEqual(response.status_code, 200)
        result_ids = [r["id"] for r in response.data["results"]]
        self.assertEqual(len(result_ids), 1)
        self.assertIn(self.newsletter_topic.id, result_ids)

        response = self.client.get(
            url, {"category": SubscriptionTopic.TopicCategory.MARKETING}
        )
        self.assertEqual(response.status_code, 200)
        result_ids = [r["id"] for r in response.data["results"]]
        self.assertEqual(len(result_ids), 1)
        self.assertIn(self.marketing_topic.id, result_ids)

    def test_boolean_filters(self):
        url = reverse("user-subscription-topic-list")

        response = self.client.get(url, {"is_active": "true"})
        self.assertEqual(response.status_code, 200)
        result_ids = [r["id"] for r in response.data["results"]]
        self.assertGreaterEqual(len(result_ids), 3)
        self.assertIn(self.newsletter_topic.id, result_ids)
        self.assertIn(self.marketing_topic.id, result_ids)
        self.assertIn(self.system_topic.id, result_ids)
        self.assertNotIn(self.product_topic.id, result_ids)

        response = self.client.get(url, {"is_active": "false"})
        self.assertEqual(response.status_code, 200)

        self.assertFalse(
            self.product_topic.is_active, "Product topic should be inactive"
        )

        response = self.client.get(url, {"is_default": "true"})
        self.assertEqual(response.status_code, 200)
        result_ids = [r["id"] for r in response.data["results"]]
        self.assertGreaterEqual(len(result_ids), 2)
        self.assertIn(self.newsletter_topic.id, result_ids)
        self.assertIn(self.system_topic.id, result_ids)

        response = self.client.get(url, {"requires_confirmation": "true"})
        self.assertEqual(response.status_code, 200)
        result_ids = [r["id"] for r in response.data["results"]]
        self.assertGreaterEqual(len(result_ids), 1)
        self.assertIn(self.marketing_topic.id, result_ids)

    def test_slug_filters(self):
        url = reverse("user-subscription-topic-list")

        response = self.client.get(url, {"slug": "newsletter"})
        self.assertEqual(response.status_code, 200)
        result_ids = [r["id"] for r in response.data["results"]]
        self.assertGreaterEqual(len(result_ids), 1)
        self.assertIn(self.newsletter_topic.id, result_ids)

        response = self.client.get(url, {"slug_exact": "weekly-newsletter"})
        self.assertEqual(response.status_code, 200)
        result_ids = [r["id"] for r in response.data["results"]]
        self.assertEqual(len(result_ids), 1)
        self.assertIn(self.newsletter_topic.id, result_ids)

        response = self.client.get(url, {"slug": "system"})
        self.assertEqual(response.status_code, 200)
        result_ids = [r["id"] for r in response.data["results"]]
        self.assertGreaterEqual(len(result_ids), 1)
        self.assertIn(self.system_topic.id, result_ids)

    def test_translation_filters(self):
        url = reverse("user-subscription-topic-list")

        response = self.client.get(url, {"name": "Newsletter"})
        self.assertEqual(response.status_code, 200)
        result_ids = [r["id"] for r in response.data["results"]]
        self.assertEqual(len(result_ids), 1)
        self.assertIn(self.newsletter_topic.id, result_ids)

        response = self.client.get(url, {"name": "Updates"})
        self.assertEqual(response.status_code, 200)
        result_ids = [r["id"] for r in response.data["results"]]
        self.assertEqual(len(result_ids), 1)
        self.assertIn(self.marketing_topic.id, result_ids)

        response = self.client.get(url, {"description": "weekly"})
        self.assertEqual(response.status_code, 200)
        result_ids = [r["id"] for r in response.data["results"]]
        self.assertEqual(len(result_ids), 1)
        self.assertIn(self.newsletter_topic.id, result_ids)

        response = self.client.get(url, {"description": "system"})
        self.assertEqual(response.status_code, 200)
        result_ids = [r["id"] for r in response.data["results"]]
        self.assertEqual(len(result_ids), 1)
        self.assertIn(self.system_topic.id, result_ids)

    def test_has_subscribers_filter(self):
        url = reverse("user-subscription-topic-list")

        UserSubscription.objects.all().delete()

        try:
            UserSubscriptionFactory(
                user=self.user1, topic=self.newsletter_topic
            )
            UserSubscriptionFactory(user=self.user1, topic=self.marketing_topic)
        except Exception:
            self.skipTest("Subscription creation failed due to constraints")

        response = self.client.get(url, {"has_subscribers": "true"})
        self.assertEqual(response.status_code, 200)
        result_ids = [r["id"] for r in response.data["results"]]
        self.assertGreaterEqual(len(result_ids), 2)
        self.assertIn(self.newsletter_topic.id, result_ids)
        self.assertIn(self.marketing_topic.id, result_ids)

        response = self.client.get(url, {"has_subscribers": "false"})
        self.assertEqual(response.status_code, 200)
        result_ids = [r["id"] for r in response.data["results"]]
        self.assertGreaterEqual(len(result_ids), 1)
        self.assertIn(self.system_topic.id, result_ids)

    def test_complex_filter_combinations(self):
        url = reverse("user-subscription-topic-list")

        created_after_date = self.now - timedelta(days=25)
        response = self.client.get(
            url,
            {
                "createdAfter": created_after_date.isoformat(),
                "isActive": "true",
                "category": SubscriptionTopic.TopicCategory.MARKETING,
                "requiresConfirmation": "true",
                "ordering": "-createdAt",
            },
        )

        self.assertEqual(response.status_code, 200)
        result_ids = [r["id"] for r in response.data["results"]]
        self.assertEqual(len(result_ids), 1)
        self.assertIn(self.marketing_topic.id, result_ids)

        response = self.client.get(
            url,
            {
                "name": "Newsletter",
                "category": SubscriptionTopic.TopicCategory.NEWSLETTER,
                "isDefault": "true",
            },
        )

        self.assertEqual(response.status_code, 200)
        result_ids = [r["id"] for r in response.data["results"]]
        self.assertEqual(len(result_ids), 1)
        self.assertIn(self.newsletter_topic.id, result_ids)

    def test_filter_with_ordering(self):
        url = reverse("user-subscription-topic-list")

        response = self.client.get(url, {"ordering": "-createdAt"})
        self.assertEqual(response.status_code, 200)

        results = response.data["results"]
        self.assertGreaterEqual(len(results), 3)

        topic_ids = [r["id"] for r in results]
        system_index = topic_ids.index(self.system_topic.id)
        newsletter_index = topic_ids.index(self.newsletter_topic.id)
        self.assertLess(
            system_index,
            newsletter_index,
            "System topic should come before newsletter topic when ordered by -createdAt",
        )

        response = self.client.get(url, {"ordering": "category,slug"})
        self.assertEqual(response.status_code, 200)

        results = response.data["results"]
        self.assertGreaterEqual(len(results), 3)
        categories = [r["category"] for r in results]
        if "MARKETING" in categories and "NEWSLETTER" in categories:
            self.assertLess(
                categories.index("MARKETING"),
                categories.index("NEWSLETTER"),
                "Marketing should come before Newsletter alphabetically",
            )

    def test_existing_filters_still_work(self):
        url = reverse("user-subscription-topic-list")

        response = self.client.get(
            url,
            {
                "is_active": "true",
                "is_default": "true",
                "category": SubscriptionTopic.TopicCategory.NEWSLETTER,
            },
        )
        self.assertEqual(response.status_code, 200)

        result_ids = [r["id"] for r in response.data["results"]]
        self.assertEqual(len(result_ids), 1)
        self.assertIn(self.newsletter_topic.id, result_ids)

    def tearDown(self):
        UserSubscription.objects.all().delete()
        SubscriptionTopic.objects.all().delete()



# Generated by Django 5.0.3 on 2024-03-31 12:31
from django.db import migrations
from django.db import models


class Migration(migrations.Migration):
    dependencies = [
        ("core", "0003_settings_site_alter_settings_unique_together"),
        ("sites", "0002_alter_domain_unique"),
    ]

    operations = [
        migrations.AddIndex(
            model_name="settings",
            index=models.Index(fields=["value_type"], name="settings_value_type_idx"),
        ),
    ]



from order.factories.history import OrderHistoryFactory, OrderItemHistoryFactory
from order.factories.item import OrderItemFactory
from order.factories.order import OrderFactory

__all__ = [
    "OrderFactory",
    "OrderHistoryFactory",
    "OrderItemFactory",
    "OrderItemHistoryFactory",
]



from unittest.mock import patch

from contact.utils import (
    detect_spam_patterns,
    sanitize_message,
    validate_contact_content,
)


class TestDetectSpamPatterns:
    def test_spam_keywords_detection(self):
        spam_message = "Congratulations! You've won the lottery! Click here for free money!"
        result = detect_spam_patterns(
            spam_message, "test@example.com", "Test User"
        )
        assert result is True

    def test_single_spam_keyword_not_spam(self):
        message = "I need help with my order discount code"
        result = detect_spam_patterns(message, "test@example.com", "Test User")
        assert result is False

    def test_multiple_spam_keywords_triggers_spam(self):
        message = "Best deal ever! Guaranteed lowest price! Act now!"
        result = detect_spam_patterns(message, "test@example.com", "Test User")
        assert result is True

    def test_long_message_spam_detection(self):
        long_message = "a" * 5001
        result = detect_spam_patterns(
            long_message, "test@example.com", "Test User"
        )
        assert result is True

    def test_multiple_links_spam_detection(self):
        message = "Check out http://site1.com and http://site2.com and http://site3.com and http://site4.com"
        result = detect_spam_patterns(message, "test@example.com", "Test User")
        assert result is True

    def test_repeated_characters_spam_detection(self):
        message = "HEEEEEEEEEEELLLLLLLLLLPPPPPPPPP MEEEEEEEEEE"
        result = detect_spam_patterns(message, "test@example.com", "Test User")
        assert result is True

    def test_very_short_message_spam_detection(self):
        message = "Hi"
        result = detect_spam_patterns(message, "test@example.com", "Test User")
        assert result is True

    def test_normal_message_not_spam(self):
        message = "Hello, I have a question about your product catalog and pricing options."
        result = detect_spam_patterns(message, "test@example.com", "Test User")
        assert result is False

    def test_case_insensitive_spam_detection(self):
        message = "VIAGRA CASINO LOTTERY winner!"
        result = detect_spam_patterns(message, "test@example.com", "Test User")
        assert result is True

    def test_exactly_three_spam_keywords_triggers_spam(self):
        message = (
            "Special promotion with guaranteed discount - limited time offer!"
        )
        result = detect_spam_patterns(message, "test@example.com", "Test User")
        assert result is True


class TestSanitizeMessage:
    def test_remove_html_tags(self):
        message = "Hello <script>alert('test')</script> world <b>bold</b> text"
        result = sanitize_message(message)
        expected = "Hello alert('test') world bold text"
        assert result == expected

    def test_reduce_repeated_characters(self):
        message = "Helllllllllllo woooooooorld"
        result = sanitize_message(message)
        expected = "Helllo wooorld"
        assert result == expected

    def test_normalize_whitespace(self):
        message = "Hello    world\n\n\nwith   multiple   spaces"
        result = sanitize_message(message)
        expected = "Hello world with multiple spaces"
        assert result == expected

    def test_strip_leading_trailing_whitespace(self):
        message = "   Hello world   "
        result = sanitize_message(message)
        expected = "Hello world"
        assert result == expected

    def test_empty_message_handling(self):
        message = ""
        result = sanitize_message(message)
        expected = ""
        assert result == expected

    def test_whitespace_only_message(self):
        message = "   \n\n\t\t   "
        result = sanitize_message(message)
        expected = ""
        assert result == expected

    def test_complex_sanitization(self):
        message = "  <div>Hellllllllllo</div>    woooooorld\n\n\nwith<script>bad</script>   multiple   spaces  "
        result = sanitize_message(message)
        expected = "Helllo wooorld withbad multiple spaces"
        assert result == expected

    def test_preserve_normal_repeated_characters(self):
        message = "Hello world with some letters like 'book' and 'cool'"
        result = sanitize_message(message)
        expected = "Hello world with some letters like 'book' and 'cool'"
        assert result == expected


class TestValidateContactContent:
    def test_valid_contact_content(self):
        result = validate_contact_content(
            name="John Doe",
            email="john@example.com",
            message="This is a valid message with sufficient length.",
        )
        assert result["valid"] is True
        assert len(result["errors"]) == 0
        assert "warnings" in result

    def test_short_name_validation(self):
        result = validate_contact_content(
            name="J",
            email="john@example.com",
            message="This is a valid message with sufficient length.",
        )
        assert result["valid"] is False
        assert "name" in result["errors"]
        assert "at least 2 characters" in str(result["errors"]["name"])

    def test_empty_name_validation(self):
        result = validate_contact_content(
            name="",
            email="john@example.com",
            message="This is a valid message with sufficient length.",
        )
        assert result["valid"] is False
        assert "name" in result["errors"]

    def test_whitespace_only_name_validation(self):
        result = validate_contact_content(
            name="  ",
            email="john@example.com",
            message="This is a valid message with sufficient length.",
        )
        assert result["valid"] is False
        assert "name" in result["errors"]

    def test_short_message_validation(self):
        result = validate_contact_content(
            name="John Doe", email="john@example.com", message="Short"
        )
        assert result["valid"] is False
        assert "message" in result["errors"]
        assert "at least 10 characters" in str(result["errors"]["message"])

    def test_long_message_validation(self):
        long_message = "a" * 5001
        result = validate_contact_content(
            name="John Doe", email="john@example.com", message=long_message
        )
        assert result["valid"] is False
        assert "message" in result["errors"]
        assert "too long" in str(result["errors"]["message"])

    def test_invalid_email_validation(self):
        result = validate_contact_content(
            name="John Doe",
            email="invalid-email",
            message="This is a valid message with sufficient length.",
        )
        assert result["valid"] is False
        assert "email" in result["errors"]
        assert "valid email" in str(result["errors"]["email"])

    def test_various_invalid_email_formats(self):
        invalid_emails = [
            "test@",
            "@example.com",
            "test@.com",
            "test.example.com",
            "test@example",
            "test@@example.com",
            "",
        ]

        for invalid_email in invalid_emails:
            result = validate_contact_content(
                name="John Doe",
                email=invalid_email,
                message="This is a valid message with sufficient length.",
            )
            assert result["valid"] is False
            assert "email" in result["errors"]

    def test_valid_email_formats(self):
        valid_emails = [
            "test@example.com",
            "user.name@example.com",
            "user+tag@example.co.uk",
            "123@example.org",
            "test-email@sub.example.com",
        ]

        for valid_email in valid_emails:
            result = validate_contact_content(
                name="John Doe",
                email=valid_email,
                message="This is a valid message with sufficient length.",
            )
            assert "email" not in result["errors"] or "valid email" not in str(
                result["errors"].get("email", "")
            )

    @patch("contact.utils.detect_spam_patterns")
    def test_spam_detection_integration(self, mock_detect_spam):
        mock_detect_spam.return_value = True

        result = validate_contact_content(
            name="John Doe",
            email="john@example.com",
            message="This is a valid message with sufficient length.",
        )

        assert result["valid"] is False
        assert "spam" in result["errors"]
        assert "appears to be spam" in str(result["errors"]["spam"])
        mock_detect_spam.assert_called_once()

    @patch("contact.utils.is_disposable_domain")
    def test_disposable_email_detection(self, mock_is_disposable):
        mock_is_disposable.return_value = True

        result = validate_contact_content(
            name="John Doe",
            email="test@tempmail.com",
            message="This is a valid message with sufficient length.",
        )

        assert result["valid"] is False
        assert "email" in result["errors"]
        assert "Disposable" in str(result["errors"]["email"])
        mock_is_disposable.assert_called_once_with("tempmail.com")

    def test_multiple_validation_errors(self):
        result = validate_contact_content(
            name="J", email="invalid-email", message="Short"
        )

        assert result["valid"] is False
        assert len(result["errors"]) >= 3
        assert "name" in result["errors"]
        assert "email" in result["errors"]
        assert "message" in result["errors"]

    def test_whitespace_message_validation(self):
        result = validate_contact_content(
            name="John Doe", email="john@example.com", message="   \n\n\t   "
        )
        assert result["valid"] is False
        assert "message" in result["errors"]

    def test_exactly_minimum_length_inputs(self):
        result = validate_contact_content(
            name="Jo",
            email="a@b.co",
            message="1234567890",
        )
        assert "name" not in result[
            "errors"
        ] or "at least 2 characters" not in str(result["errors"]["name"])
        assert "message" not in result[
            "errors"
        ] or "at least 10 characters" not in str(result["errors"]["message"])



from django.contrib.auth import get_user_model
from django.utils.translation import gettext_lazy as _
from drf_spectacular.utils import extend_schema_field
from rest_framework import serializers
from rest_framework.relations import PrimaryKeyRelatedField

from core.api.schema import generate_schema_multi_lang
from core.utils.serializers import TranslatedFieldExtended
from product.models.favourite import ProductFavourite
from product.models.product import Product
from product.serializers.product import (
    ProductDetailSerializer,
)

User = get_user_model()


@extend_schema_field(generate_schema_multi_lang(Product))
class TranslatedFieldsFieldExtend(TranslatedFieldExtended):
    pass


class ProductFavouriteSerializer(serializers.ModelSerializer[ProductFavourite]):
    user_id = serializers.IntegerField(source="user.id", read_only=True)
    user_username = serializers.CharField(
        source="user.username", read_only=True
    )
    product = ProductDetailSerializer(read_only=True)

    class Meta:
        model = ProductFavourite
        fields = (
            "id",
            "user_id",
            "user_username",
            "product",
            "created_at",
            "uuid",
        )


class ProductFavouriteDetailSerializer(ProductFavouriteSerializer):
    user = serializers.StringRelatedField(read_only=True)

    class Meta(ProductFavouriteSerializer.Meta):
        fields = (
            *ProductFavouriteSerializer.Meta.fields,
            "user",
            "updated_at",
        )


class ProductFavouriteWriteSerializer(
    serializers.ModelSerializer[ProductFavourite]
):
    product = PrimaryKeyRelatedField(queryset=Product.objects.all(), many=False)

    class Meta:
        model = ProductFavourite
        fields = (
            "id",
            "product",
        )
        read_only_fields = ("id",)

    def validate(self, attrs):
        user = self.context["request"].user
        product = attrs.get("product")

        if (
            not self.instance
            and ProductFavourite.objects.filter(
                user=user, product=product
            ).exists()
        ):
            raise serializers.ValidationError(
                _("Product is already in favorites")
            )

        return attrs


class ProductFavouriteByProductsRequestSerializer(serializers.Serializer):
    product_ids = serializers.ListField(
        child=serializers.IntegerField(),
        help_text=_("List of product IDs to check for favorites"),
    )


class ProductFavouriteByProductsResponseSerializer(serializers.Serializer):
    id = serializers.IntegerField()
    user_id = serializers.IntegerField()
    product_id = serializers.IntegerField()
    created_at = serializers.DateTimeField()


class ProductDetailResponseSerializer(ProductDetailSerializer):
    class Meta(ProductDetailSerializer.Meta):
        pass



# Generated by Django 5.0.4 on 2024-04-24 09:29
import django.contrib.postgres.search
from django.db import migrations
from django.db import models


class Migration(migrations.Migration):
    dependencies = [
        ("blog", "0012_alter_blogcomment_likes_alter_blogpost_likes"),
    ]

    operations = [
        migrations.AddField(
            model_name="blogposttranslation",
            name="search_document",
            field=models.TextField(
                blank=True,
                db_index=True,
                default="",
                verbose_name="Search Document",
            ),
        ),
        migrations.AddField(
            model_name="blogposttranslation",
            name="search_document_dirty",
            field=models.BooleanField(
                db_index=True,
                default=False,
                verbose_name="Search Document Dirty",
            ),
        ),
        migrations.AddField(
            model_name="blogposttranslation",
            name="search_vector",
            field=django.contrib.postgres.search.SearchVectorField(
                blank=True,
                db_index=True,
                null=True,
                verbose_name="Search Vector",
            ),
        ),
        migrations.AddField(
            model_name="blogposttranslation",
            name="search_vector_dirty",
            field=models.BooleanField(
                db_index=True,
                default=False,
                verbose_name="Search Vector Dirty",
            ),
        ),
    ]



# Generated by Django 4.2.9 on 2024-02-10 14:39
import django.db.models.deletion
import mptt.fields
import parler.fields
from django.conf import settings
from django.db import migrations
from django.db import models


class Migration(migrations.Migration):
    initial = True

    dependencies = [
        migrations.swappable_dependency(settings.AUTH_USER_MODEL),
        ("blog", "0001_initial"),
    ]

    operations = [
        migrations.AddField(
            model_name="blogpost",
            name="likes",
            field=models.ManyToManyField(
                blank=True,
                related_name="blog_post_likes",
                to=settings.AUTH_USER_MODEL,
            ),
        ),
        migrations.AddField(
            model_name="blogpost",
            name="tags",
            field=models.ManyToManyField(
                blank=True, related_name="blog_post_tags", to="blog.blogtag"
            ),
        ),
        migrations.AddField(
            model_name="blogcommenttranslation",
            name="master",
            field=parler.fields.TranslationsForeignKey(
                editable=False,
                null=True,
                on_delete=django.db.models.deletion.CASCADE,
                related_name="translations",
                to="blog.blogcomment",
            ),
        ),
        migrations.AddField(
            model_name="blogcomment",
            name="likes",
            field=models.ManyToManyField(
                blank=True,
                related_name="blog_comment_likes",
                to=settings.AUTH_USER_MODEL,
            ),
        ),
        migrations.AddField(
            model_name="blogcomment",
            name="post",
            field=models.ForeignKey(
                null=True,
                on_delete=django.db.models.deletion.SET_NULL,
                related_name="blog_comment_post",
                to="blog.blogpost",
            ),
        ),
        migrations.AddField(
            model_name="blogcomment",
            name="user",
            field=models.ForeignKey(
                null=True,
                on_delete=django.db.models.deletion.SET_NULL,
                related_name="blog_comment_user",
                to=settings.AUTH_USER_MODEL,
            ),
        ),
        migrations.AddField(
            model_name="blogcategorytranslation",
            name="master",
            field=parler.fields.TranslationsForeignKey(
                editable=False,
                null=True,
                on_delete=django.db.models.deletion.CASCADE,
                related_name="translations",
                to="blog.blogcategory",
            ),
        ),
        migrations.AddField(
            model_name="blogcategory",
            name="parent",
            field=mptt.fields.TreeForeignKey(
                blank=True,
                null=True,
                on_delete=django.db.models.deletion.CASCADE,
                related_name="children",
                to="blog.blogcategory",
            ),
        ),
        migrations.AddField(
            model_name="blogauthortranslation",
            name="master",
            field=parler.fields.TranslationsForeignKey(
                editable=False,
                null=True,
                on_delete=django.db.models.deletion.CASCADE,
                related_name="translations",
                to="blog.blogauthor",
            ),
        ),
        migrations.AddField(
            model_name="blogauthor",
            name="user",
            field=models.OneToOneField(
                on_delete=django.db.models.deletion.PROTECT,
                to=settings.AUTH_USER_MODEL,
            ),
        ),
        migrations.AlterUniqueTogether(
            name="blogtagtranslation",
            unique_together={("language_code", "master")},
        ),
        migrations.AlterUniqueTogether(
            name="blogposttranslation",
            unique_together={("language_code", "master")},
        ),
        migrations.AlterUniqueTogether(
            name="blogcommenttranslation",
            unique_together={("language_code", "master")},
        ),
        migrations.AddConstraint(
            model_name="blogcomment",
            constraint=models.UniqueConstraint(
                fields=("user", "post"), name="unique_blog_comment"
            ),
        ),
        migrations.AlterUniqueTogether(
            name="blogcategorytranslation",
            unique_together={("language_code", "master")},
        ),
        migrations.AlterUniqueTogether(
            name="blogauthortranslation",
            unique_together={("language_code", "master")},
        ),
    ]



# Generated by Django 4.2.9 on 2024-02-14 21:13
import django.db.models.deletion
import mptt.fields
from django.conf import settings
from django.db import migrations
from django.db import models


class Migration(migrations.Migration):
    dependencies = [
        migrations.swappable_dependency(settings.AUTH_USER_MODEL),
        ("blog", "0002_initial"),
    ]

    operations = [
        migrations.AddField(
            model_name="blogcomment",
            name="level",
            field=models.PositiveIntegerField(default=0, editable=False),
            preserve_default=False,
        ),
        migrations.AddField(
            model_name="blogcomment",
            name="lft",
            field=models.PositiveIntegerField(default=1, editable=False),
            preserve_default=False,
        ),
        migrations.AddField(
            model_name="blogcomment",
            name="parent",
            field=mptt.fields.TreeForeignKey(
                blank=True,
                null=True,
                on_delete=django.db.models.deletion.CASCADE,
                related_name="children",
                to="blog.blogcomment",
            ),
        ),
        migrations.AddField(
            model_name="blogcomment",
            name="rght",
            field=models.PositiveIntegerField(default=2, editable=False),
            preserve_default=False,
        ),
        migrations.AddField(
            model_name="blogcomment",
            name="tree_id",
            field=models.PositiveIntegerField(db_index=True, default=1, editable=False),
            preserve_default=False,
        ),
        migrations.AlterField(
            model_name="blogcomment",
            name="post",
            field=models.ForeignKey(
                blank=True,
                null=True,
                on_delete=django.db.models.deletion.SET_NULL,
                related_name="blog_comment_post",
                to="blog.blogpost",
            ),
        ),
        migrations.AlterField(
            model_name="blogcomment",
            name="user",
            field=models.ForeignKey(
                blank=True,
                null=True,
                on_delete=django.db.models.deletion.SET_NULL,
                related_name="blog_comment_user",
                to=settings.AUTH_USER_MODEL,
            ),
        ),
    ]



import csv
import xml.etree.ElementTree as ET
from io import StringIO
from unittest.mock import MagicMock

from django.conf import settings
from django.contrib.messages.storage.fallback import FallbackStorage
from django.db import models
from django.db.models.options import Options
from django.http import HttpRequest
from django.test import TestCase, override_settings
from djmoney.models.fields import Money
from parler.models import TranslatableModel

from core.admin import ExportActionMixin


class MockAdminModel(models.Model):
    name = models.CharField(max_length=100)
    price = models.DecimalField(max_digits=10, decimal_places=2)
    description = models.TextField()

    class Meta:
        app_label = "test_app"
        verbose_name = "Mock Admin Model"

    def __str__(self):
        return self.name


class MockTranslatableModel(TranslatableModel):
    code = models.CharField(max_length=50)

    class Meta:
        app_label = "test_app"
        verbose_name = "test_translatable_model"

    class TranslatableMeta:
        fields = ["title", "description"]


class MockParlerMeta:
    @staticmethod
    def get_translated_fields():
        return ["title", "description"]


class ExportActionMixinTest(TestCase):
    def setUp(self):
        self.mixin = ExportActionMixin()
        self.mixin.model = MockAdminModel

        self.request = HttpRequest()
        self.request.session = "session"
        messages = FallbackStorage(self.request)
        self.request._messages = messages

        self.queryset = MagicMock()

        self.test_model_instance = MagicMock(spec=MockAdminModel)
        self.test_model_instance.name = "Test Model"
        self.test_model_instance.price = Money(
            amount=10.50, currency=settings.DEFAULT_CURRENCY
        )
        self.test_model_instance.description = "Test description"

        MockTranslatableModel._parler_meta = MockParlerMeta()
        self.translatable_model_instance = MagicMock(spec=MockTranslatableModel)
        self.translatable_model_instance.code = "test-code"
        self.en_translation = MagicMock()
        self.en_translation.title = "English Title"
        self.en_translation.description = "English Description"
        self.fr_translation = MagicMock()
        self.fr_translation.title = "French Title"
        self.fr_translation.description = "French Description"

        def mock_get_translation(lang_code):
            if lang_code == "en":
                return self.en_translation
            elif lang_code == "fr":
                return self.fr_translation
            raise models.ObjectDoesNotExist

        self.translatable_model_instance.get_translation = mock_get_translation

    def test_get_exportable_fields(self):
        mock_opts = MagicMock(spec=Options)

        regular_field = MagicMock()
        regular_field.many_to_many = False
        regular_field.one_to_many = False
        regular_field.one_to_one = False

        m2m_field = MagicMock()
        m2m_field.many_to_many = True
        m2m_field.one_to_many = False
        m2m_field.one_to_one = False

        o2m_field = MagicMock()
        o2m_field.many_to_many = False
        o2m_field.one_to_many = True
        o2m_field.one_to_one = False

        o2o_field = MagicMock()
        o2o_field.many_to_many = False
        o2o_field.one_to_many = False
        o2o_field.one_to_one = True

        mock_opts.get_fields.return_value = [
            regular_field,
            m2m_field,
            o2m_field,
            o2o_field,
        ]

        fields = self.mixin._get_exportable_fields(mock_opts)

        self.assertEqual(len(fields), 1)
        self.assertEqual(fields[0], regular_field)

    @override_settings(LANGUAGES=[("en", "English"), ("fr", "French")])
    def test_export_csv_standard_model(self):
        self.mixin.model._meta = MagicMock()
        self.mixin.model._meta.verbose_name = "test_model"

        name_field = MagicMock()
        name_field.name = "name"
        name_field.verbose_name = "Name"
        name_field.many_to_many = False
        name_field.one_to_many = False
        name_field.one_to_one = False

        price_field = MagicMock()
        price_field.name = "price"
        price_field.verbose_name = "Price"
        price_field.many_to_many = False
        price_field.one_to_many = False
        price_field.one_to_one = False

        desc_field = MagicMock()
        desc_field.name = "description"
        desc_field.verbose_name = "Description"
        desc_field.many_to_many = False
        desc_field.one_to_many = False
        desc_field.one_to_one = False

        self.mixin.model._meta.get_fields.return_value = [
            name_field,
            price_field,
            desc_field,
        ]

        self.queryset.__iter__.return_value = [self.test_model_instance]

        response = self.mixin.export_csv(self.request, self.queryset)

        self.assertEqual(response["Content-Type"], "text/csv; charset=utf-8")
        self.assertEqual(
            response["Content-Disposition"],
            "attachment; filename=test_model.csv",
        )

        content = response.content.decode("utf-8")
        reader = csv.reader(StringIO(content))
        rows = list(reader)

        self.assertEqual(rows[0], ["Name", "Price", "Description"])

        self.assertEqual(rows[1][0], "Test Model")
        self.assertEqual(rows[1][1], "10.5")
        self.assertEqual(rows[1][2], "Test description")

    @override_settings(LANGUAGES=[("en", "English"), ("fr", "French")])
    def test_export_csv_translatable_model(self):
        self.mixin.model = MockTranslatableModel
        self.mixin.model._meta = MagicMock()
        self.mixin.model._meta.verbose_name = "test_translatable_model"

        code_field = MagicMock()
        code_field.name = "code"
        code_field.verbose_name = "Code"
        code_field.many_to_many = False
        code_field.one_to_many = False
        code_field.one_to_one = False

        self.mixin.model._meta.get_fields.return_value = [code_field]

        self.queryset.__iter__.return_value = [self.translatable_model_instance]

        response = self.mixin.export_csv(self.request, self.queryset)

        content = response.content.decode("utf-8")
        reader = csv.reader(StringIO(content))
        rows = list(reader)

        expected_headers = [
            "Code",
            "title_en",
            "description_en",
            "title_fr",
            "description_fr",
        ]
        self.assertEqual(rows[0], expected_headers)

        self.assertEqual(rows[1][0], "test-code")
        self.assertEqual(rows[1][1], "English Title")
        self.assertEqual(rows[1][2], "English Description")
        self.assertEqual(rows[1][3], "French Title")
        self.assertEqual(rows[1][4], "French Description")

    def test_add_base_fields_to_xml(self):
        root = ET.Element("root")
        obj_element = ET.SubElement(root, "test_model")

        name_field = MagicMock()
        name_field.name = "name"

        price_field = MagicMock()
        price_field.name = "price"

        json_field = MagicMock(spec=models.JSONField)
        json_field.name = "json_data"
        self.test_model_instance.json_data = {"key": "value"}

        self.mixin._add_base_fields_to_xml(
            self.test_model_instance,
            obj_element,
            [name_field, price_field, json_field],
        )

        self.assertEqual(obj_element.find("name").text, "Test Model")
        self.assertEqual(obj_element.find("price").text, "10.5")
        self.assertEqual(obj_element.find("json_data").text, '{"key": "value"}')

    @override_settings(LANGUAGES=[("en", "English"), ("fr", "French")])
    def test_add_translated_fields_to_xml(self):
        root = ET.Element("root")
        obj_element = ET.SubElement(root, "test_translatable_model")

        self.mixin._add_translated_fields_to_xml(
            self.translatable_model_instance,
            obj_element,
            ["title", "description"],
        )

        translations = obj_element.find("translations")
        self.assertIsNotNone(translations)

        en_trans = translations.find("./translation[@lang='en']")
        self.assertIsNotNone(en_trans)
        self.assertEqual(en_trans.find("title").text, "English Title")
        self.assertEqual(
            en_trans.find("description").text, "English Description"
        )

        fr_trans = translations.find("./translation[@lang='fr']")
        self.assertIsNotNone(fr_trans)
        self.assertEqual(fr_trans.find("title").text, "French Title")
        self.assertEqual(
            fr_trans.find("description").text, "French Description"
        )

    def test_generate_xml_response(self):
        root = ET.Element("test_models")
        model = ET.SubElement(root, "test_model")
        name = ET.SubElement(model, "name")
        name.text = "Test Model"

        response = self.mixin._generate_xml_response(root, "test_model")

        self.assertEqual(
            response["Content-Type"], "application/xml; charset=utf-8"
        )
        self.assertEqual(
            response["Content-Disposition"],
            "attachment; filename=test_model.xml",
        )

        content = response.content.decode("utf-8")
        self.assertTrue(
            content.startswith('<?xml version="1.0" encoding="UTF-8"?>')
        )

        parsed_root = ET.fromstring(content)
        self.assertEqual(parsed_root.tag, "test_models")
        self.assertEqual(parsed_root[0].tag, "test_model")
        self.assertEqual(parsed_root[0][0].tag, "name")
        self.assertEqual(parsed_root[0][0].text, "Test Model")

    def test_generate_xml_response_fallback(self):
        root = ET.Element("root")
        test_element = ET.SubElement(root, "test")
        test_element.text = "test content"

        response = self.mixin._generate_xml_response(root, "test")

        self.assertEqual(
            response["Content-Type"], "application/xml; charset=utf-8"
        )
        self.assertEqual(
            response["Content-Disposition"], "attachment; filename=test.xml"
        )
        self.assertIn("test content", response.content.decode("utf-8"))

    @override_settings(LANGUAGES=[("en", "English"), ("fr", "French")])
    def test_export_xml_standard_model(self):
        self.mixin.model._meta = MagicMock()
        self.mixin.model._meta.verbose_name = "test_model"

        name_field = MagicMock()
        name_field.name = "name"
        name_field.verbose_name = "Name"
        name_field.many_to_many = False
        name_field.one_to_many = False
        name_field.one_to_one = False

        self.mixin.model._meta.get_fields.return_value = [name_field]

        self.queryset.__iter__.return_value = [self.test_model_instance]
        self.queryset.__bool__.return_value = True

        response = self.mixin.export_xml(self.request, self.queryset)

        self.assertEqual(
            response["Content-Type"], "application/xml; charset=utf-8"
        )
        self.assertEqual(
            response["Content-Disposition"],
            "attachment; filename=test_model.xml",
        )

        content = response.content.decode("utf-8")
        self.assertTrue(
            content.startswith('<?xml version="1.0" encoding="UTF-8"?>')
        )

        parsed_root = ET.fromstring(content)
        self.assertEqual(parsed_root.tag, "test_models")
        self.assertEqual(parsed_root[0].tag, "test_model")
        self.assertEqual(parsed_root[0][0].tag, "name")
        self.assertEqual(parsed_root[0][0].text, "Test Model")

    @override_settings(LANGUAGES=[("en", "English"), ("fr", "French")])
    def test_export_xml_translatable_model(self):
        self.mixin.model = MockTranslatableModel
        self.mixin.model._meta = MagicMock()
        self.mixin.model._meta.verbose_name = "test_translatable_model"

        code_field = MagicMock()
        code_field.name = "code"
        code_field.verbose_name = "Code"
        code_field.many_to_many = False
        code_field.one_to_many = False
        code_field.one_to_one = False

        self.mixin.model._meta.get_fields.return_value = [code_field]

        self.queryset.__iter__.return_value = [self.translatable_model_instance]
        self.queryset.__bool__.return_value = True

        response = self.mixin.export_xml(self.request, self.queryset)

        self.assertEqual(
            response["Content-Type"], "application/xml; charset=utf-8"
        )

        content = response.content.decode("utf-8")
        parsed_root = ET.fromstring(content)

        model_element = parsed_root.find("./test_translatable_model")
        self.assertIsNotNone(model_element)
        self.assertEqual(model_element.find("./code").text, "test-code")

        translations = model_element.find("./translations")
        self.assertIsNotNone(translations)

        en_trans = translations.find("./translation[@lang='en']")
        self.assertIsNotNone(en_trans)
        self.assertEqual(en_trans.find("./title").text, "English Title")

    def test_export_xml_empty_queryset(self):
        self.queryset.__bool__.return_value = False

        response = self.mixin.export_xml(self.request, self.queryset)

        self.assertEqual(response.status_code, 400)

    def test_export_xml_error_handling(self):
        self.mixin.model._meta = MagicMock()
        self.mixin.model._meta.verbose_name = "test_model"
        self.mixin.model._meta.get_fields.return_value = []

        self.queryset.__iter__.return_value = []

        response = self.mixin.export_xml(self.request, self.queryset)

        self.assertEqual(
            response["Content-Type"], "application/xml; charset=utf-8"
        )
        self.assertEqual(
            response["Content-Disposition"],
            "attachment; filename=test_model.xml",
        )

    def test_get_export_formats(self):
        formats = self.mixin.get_export_formats()

        self.assertEqual(len(formats), 2)
        self.assertEqual(formats[0]["format"], "csv")
        self.assertEqual(formats[0]["label"], "CSV")
        self.assertEqual(formats[1]["format"], "xml")
        self.assertEqual(formats[1]["label"], "XML")



# Generated by Django 5.2.1 on 2025-05-14 20:33

import django.contrib.postgres.indexes
from django.conf import settings
from django.db import migrations


class Migration(migrations.Migration):

    dependencies = [
        ('cart', '0009_alter_cart_session_key'),
        ('product', '0022_alter_productfavourite_options_and_more'),
        migrations.swappable_dependency(settings.AUTH_USER_MODEL),
    ]

    operations = [
        migrations.AddIndex(
            model_name='cart',
            index=django.contrib.postgres.indexes.BTreeIndex(fields=['created_at'], name='cart_created_at_ix'),
        ),
        migrations.AddIndex(
            model_name='cart',
            index=django.contrib.postgres.indexes.BTreeIndex(fields=['updated_at'], name='cart_updated_at_ix'),
        ),
        migrations.AddIndex(
            model_name='cart',
            index=django.contrib.postgres.indexes.BTreeIndex(fields=['user'], name='cart_user_ix'),
        ),
        migrations.AddIndex(
            model_name='cart',
            index=django.contrib.postgres.indexes.BTreeIndex(fields=['session_key'], name='cart_session_key_ix'),
        ),
        migrations.AddIndex(
            model_name='cart',
            index=django.contrib.postgres.indexes.BTreeIndex(fields=['last_activity'], name='cart_last_activity_ix'),
        ),
        migrations.AddIndex(
            model_name='cartitem',
            index=django.contrib.postgres.indexes.BTreeIndex(fields=['created_at'], name='cartitem_created_at_ix'),
        ),
        migrations.AddIndex(
            model_name='cartitem',
            index=django.contrib.postgres.indexes.BTreeIndex(fields=['updated_at'], name='cartitem_updated_at_ix'),
        ),
        migrations.AddIndex(
            model_name='cartitem',
            index=django.contrib.postgres.indexes.BTreeIndex(fields=['cart'], name='cart_item_cart_ix'),
        ),
        migrations.AddIndex(
            model_name='cartitem',
            index=django.contrib.postgres.indexes.BTreeIndex(fields=['product'], name='cart_item_product_ix'),
        ),
        migrations.AddIndex(
            model_name='cartitem',
            index=django.contrib.postgres.indexes.BTreeIndex(fields=['quantity'], name='cart_item_quantity_ix'),
        ),
    ]
